import unittest
from gdshelpers.geometry.ebl_frame_generators import raith_marker_frame


class MarkerFrameTestCase(unittest.TestCase):
    def test_raith_marker_frame(self):
        bounds = (0, 0, 1000, 1000)

        markers = raith_marker_frame(bounds, padding=50, pitch=50, size=10, n=0)
        self.assertEqual(len(markers), 4)

        markers = raith_marker_frame(bounds, padding=50, pitch=30, size=10, n=1)
        self.assertEqual(len(markers), 12)
        markerpositions = set([tuple(m.origin) for m in markers])
        expected_markerpositions = {(-50, -50), (-50, -20), (-20, -50), (1050, 1050), (1050, 1020), (1020, 1050),
                                    (-50, 1050), (-50, 1020), (-20, 1050), (1050, -50), (1050, -20), (1020, -50)}

        self.assertEqual(markerpositions, expected_markerpositions)



import unittest
import numpy as np

from gdshelpers.parts.waveguide import Waveguide
from gdshelpers.helpers.positive_resist import convert_to_positive_resist


class PositiveResistTestCase(unittest.TestCase):
    def test_positive_resist(self):
        distance = 1e-3

        waveguide = Waveguide([0, 0], 0, 1)
        for i_bend in range(9):
            waveguide.add_bend(angle=np.pi, radius=60 + i_bend * 40)

        waveguide_positive = convert_to_positive_resist(waveguide, 1)

        self.assertFalse(waveguide.get_shapely_object().buffer(-distance).intersects(waveguide_positive))
        self.assertTrue(waveguide.get_shapely_object().buffer(distance).intersects(waveguide_positive))
        self.assertTrue(waveguide_positive.convex_hull.contains(waveguide.get_shapely_object()))



from __future__ import division, print_function

import warnings
import itertools

import numpy as np
import shapely.topology
import shapely.ops
import shapely.geometry
import shapely.geos
import shapely.validation

import gdshelpers
from gdshelpers.helpers import raith_eline_dosefactor_to_datatype


def shapely_collection_to_basic_objs(collection):
    """
    Convert the object or the collection to a list of basic shapely objects.

    :param collection:
    :return: :rtype:
    """

    if isinstance(collection, (tuple, list)):
        objs = list()

        for col in collection:
            objs.extend(shapely_collection_to_basic_objs(col))
    else:
        objs = hasattr(collection, 'geoms') and collection.geoms or [collection, ]
    return objs


def cut_shapely_object(obj, x_axis=None, y_axis=None, other_side=False):
    """
    Cut a shapely object into two halves.

    If both x_axis and y_axis are None, they are assumed to be
    the center of the bounding box.

    If only one of the axis is given, and the other axis is None,
    the object is cut at the given axis.

    If both values are given, or inferred to be the center of the bounding
    box due to being both None, the object is cut along its longer side.

    :param obj: Basically any shapely object
    :type obj: shapely.base.BaseGeometry
    :param x_axis: x-axis cut value.
    :type x_axis: float, None
    :param y_axis: y-axis cut value.
    :param other_side: If the cut axis is determined automatically, use the other direction.
    :return: A tuple of two shapely objects.
    :rtype: list
    """
    _safety_overlap = 1.
    bbox = np.array(obj.bounds).reshape(2, -1)
    bbox += ((-_safety_overlap, -_safety_overlap),
             (_safety_overlap, _safety_overlap))

    if (x_axis is None) and (y_axis is None):
        cut_center = bbox.mean(0)
    else:
        cut_center = x_axis, y_axis

    if not ((x_axis is None) ^ (y_axis is None)):
        if ((bbox[1] - bbox[0]).argmax() == 0) != other_side:
            # X-Direction is largest, cut vertically
            x_axis = cut_center[0]
            y_axis = None
        else:
            # Y-Direction is largest, cut horizontally
            x_axis = None
            y_axis = cut_center[1]

    if x_axis is not None:
        cut_boxes = [shapely.geometry.box(bbox[0][0], bbox[0][1], x_axis, bbox[1][1]),
                     shapely.geometry.box(x_axis, bbox[0][1], bbox[1][0], bbox[1][1])]
    else:
        cut_boxes = [shapely.geometry.box(bbox[0][0], bbox[0][1], bbox[1][0], y_axis),
                     shapely.geometry.box(bbox[0][0], y_axis, bbox[1][0], bbox[1][1])]

    try:
        out_polygons = list()
        for cut_box in cut_boxes:
            cut_polygon = obj.intersection(cut_box)
            out_polygons.extend(hasattr(cut_polygon, 'geoms') and list(cut_polygon.geoms) or [cut_polygon, ])
    except shapely.geos.TopologicalError:
        # In case of invalid geometries, try to fix them and refracture
        warnings.warn('Trying to cut an invalid polygon (possibly during gdsCAD conversion).')
        warnings.warn('We\'ll try to fix the polygon for you now, but try and figure out what\'s wrong, please.')
        return cut_shapely_object(obj.buffer(0), x_axis, y_axis, other_side=other_side)

    # If the original object was a polygon we may not create lines!
    if isinstance(obj, shapely.geometry.Polygon):
        out_polygons = [shape for shape in out_polygons if not isinstance(shape, (shapely.geometry.LinearRing,
                                                                                  shapely.geometry.LineString,))]

    return out_polygons


def _number_of_points(poly):
    """
    Returns the number of points in a Shapely geometry.

    :param poly: Shapely geometry object.
    :return: Number of points int the Shapely object.
    :rtype: int
    """
    if hasattr(poly, '_gdsh_n_points'):
        return poly._gdsh_n_points

    elif hasattr(poly, 'geoms'):
        poly._gdsh_n_points = sum(_number_of_points(p) for p in poly.geoms)

    elif type(poly) == shapely.geometry.Polygon:
        poly._gdsh_n_points = len(poly.exterior.coords) + sum([len(shapely.geometry.LinearRing(x).coords) for x
                                                               in list(poly.interiors)])
    else:
        poly._gdsh_n_points = len(poly.coords)

    return poly._gdsh_n_points


def heal(objs, max_points, max_interior=0):
    """
    Heal a list of Shapely geometries.

    All elements touching each other will be merged as long as the number of points remains below *max_points*.

    :param objs: List Shapely geometries.
    :type objs: list, tuple
    :param max_points: Maximum number of points.
    :type max_points: int
    :return: List of healed Shapely geometries.
    :rtype: list
    """

    segments_joined = True
    while segments_joined:
        segments_joined = False
        for i, polygon in enumerate(objs):
            if type(polygon) != shapely.geometry.Polygon:
                continue

            polygon_points = _number_of_points(polygon)

            for j, polygon2 in enumerate(objs):
                if polygon is polygon2 or polygon == polygon2:
                    continue
                if type(polygon2) != shapely.geometry.Polygon:
                    continue

                if max_points and polygon_points + _number_of_points(polygon2) >= (max_points - 2):
                    continue

                if polygon.touches(polygon2):
                    joined_poly = polygon.union(polygon2)

                    if type(joined_poly) != shapely.geometry.Polygon:
                        continue

                    if len(joined_poly.interiors) > max_interior:
                        continue

                    if max_points and _number_of_points(joined_poly) >= max_points:
                        continue

                    objs.append(joined_poly)
                    segments_joined = True
                    if i > j:
                        del objs[i]
                        del objs[j]
                    else:
                        del objs[j]
                        del objs[i]

                    break
    return objs


def fracture(obj, max_points_poly, max_points_line, max_interior=0):
    out_polygons = [obj, ]

    done_counter = 0
    while done_counter < len(out_polygons):
        current_polygon = out_polygons[done_counter]

        max_points = max_points_poly if type(current_polygon) == shapely.geometry.Polygon else max_points_line

        if type(current_polygon) == shapely.geometry.Polygon and len(current_polygon.interiors) > max_interior:
            uncut_polygon = out_polygons.pop(done_counter)
            cut_point = current_polygon.interiors[0].centroid.coords[0]
            shapes = cut_shapely_object(uncut_polygon, cut_point[0], cut_point[1])
            extended_shapes = shapely_collection_to_basic_objs(shapes)
            out_polygons.extend(extended_shapes)

        elif max_points and _number_of_points(current_polygon) > max_points > 0:
            uncut_polygon = out_polygons.pop(done_counter)
            shapes = cut_shapely_object(uncut_polygon)
            extended_shapes = shapely_collection_to_basic_objs(shapes)
            out_polygons.extend(extended_shapes)
        else:
            done_counter += 1

    return out_polygons


def fracture_intelligently(obj, max_points, max_points_line, over_fracture_factor=1):
    if over_fracture_factor >= 1:
        fractured_obj = fracture(obj, max_points / over_fracture_factor, max_points_line)

        # Now that the polygons have been fractured, we might be able to join them back
        # into bigger segments.
        return heal(fractured_obj, max_points)
    else:
        return fracture(obj, max_points, max_points_line)


def geometric_union(objs):
    """
    Join a list of Parts and/or Shapely objects to one big Shapely object.

    :param objs: List of Parts and Shapely objects.
    :return: Merged Shapely geometry.
    :rtype: shapely.base.BaseGeometry
    """
    objs = [obj.get_shapely_object() if hasattr(obj, 'get_shapely_object') else obj for obj in objs]
    return shapely.ops.cascaded_union(objs)


def convert_to_gdscad(objs, layer=1, datatype=None, path_width=1.0, path_pathtype=0, max_points=None,
                      over_fracture_factor=1, max_points_line=None):
    """
    Convert any shapely or list of shapely objects to a list of gdsCAD objects.

    Since Shapely objects do not contain layer information nor line width for Shapely lines, you have to specify
    these during conversion. Typically, you will only convert polygons and specify the *layer*.

    Export options such as *datatype* and maximum number of points per Polygon/Line default to the current module
    default.

    On special feature is the *over_fracture_factor*. The polygons will be fractured into
    *max_points*/*over_fracture_factor* points and then healed again - which results in better fragmentation of some
    geometries. An *over_fracture_factor* of 0 will suppress healing.

    :param objs: Part or Shapely object or list of Parts and/or Shapely objects.
    :param layer: Layer on which to put the objects.
    :type layer: int
    :param datatype: GDS datatype of the converted objects. Defaults to module wide settings.
    :type datatype: int
    :param path_width: With of GDS path, converted from Shapely lines.
    :type path_width: float
    :param path_pathtype: GDS path end type.
    :type path_pathtype: int
    :param max_points: Maximum number of points. Defaults to module wide settings.
    :type max_points: int, None
    :param max_points_line: Maximum number of points for lines. Defaults to module wide settings.
    :type max_points_line: int, None
    :param over_fracture_factor: Break polygons in *over_fracture_factor* times smaller objects first. Then merge them
                                 again. May result in better fracturing but high numbers increase conversion time.

                                 When *over_fracture_factor* is set to 0, no additional healing is done.
    :type over_fracture_factor: int
    """

    return convert_to_layout_objs(objs, layer, datatype, path_width, path_pathtype, max_points, over_fracture_factor,
                                  max_points_line, library='gdscad')


def convert_to_layout_objs(objs, layer=1, datatype=None, path_width=1.0, path_pathtype=0, max_points=None,
                           over_fracture_factor=1, max_points_line=None, library='gdscad', grid_steps_per_micron=1000):
    """
    Convert any shapely or list of shapely objects to a list of gdsCAD objects.

    Since Shapely objects do not contain layer information nor line width for Shapely lines, you have to specify
    these during conversion. Typically, you will only convert polygons and specify the *layer*.

    Export options such as *datatype* and maximum number of points per Polygon/Line default to the current module
    default.

    On special feature is the *over_fracture_factor*. The polygons will be fractured into
    *max_points*/*over_fracture_factor* points and then healed again - which results in better fragmentation of some
    geometries. An *over_fracture_factor* of 0 will suppress healing.

    :param objs: Part or Shapely object or list of Parts and/or Shapely objects.
    :param layer: Layer on which to put the objects.
    :type layer: int
    :param datatype: GDS datatype of the converted objects. Defaults to module wide settings.
    :type datatype: int
    :param path_width: With of GDS path, converted from Shapely lines.
    :type path_width: float
    :param path_pathtype: GDS path end type.
    :type path_pathtype: int
    :param max_points: Maximum number of points. Defaults to module wide settings.
    :type max_points: int, None
    :param max_points_line: Maximum number of points for lines. Defaults to module wide settings.
    :type max_points_line: int, None
    :param over_fracture_factor: Break polygons in *over_fracture_factor* times smaller objects first. Then merge them
                                 again. May result in better fracturing but high numbers increase conversion time.

                                 When *over_fracture_factor* is set to 0, no additional healing is done.
    :type over_fracture_factor: int
    :param library: Defines the used library, either gdscad or gdspy
    :type library: str
    :param grid_steps_per_micron: Number of steps of the grid per micron, defaults to 1000 steps per micron
    :type grid_steps_per_micron: int
    """
    if library == 'gdscad':
        import gdsCAD
    elif library == 'gdspy':
        import gdspy
    elif library == 'oasis':
        import fatamorgana.records
    else:
        raise AssertionError('library must be either "gdscad" or "gdspy"')

    # If number of maximum points is not specified, default to the current profile default
    max_points = max_points if max_points is not None else gdshelpers.configuration.point_limit
    max_points_line = max_points_line if max_points_line is not None else gdshelpers.configuration.point_limit_line

    # If datatype is not specified, set it the same as layer
    if datatype is None:
        policy = gdshelpers.configuration.datatype_policy
        if policy == gdshelpers.configuration.DefaultDatatype.zero:
            datatype = 0
        elif policy == gdshelpers.configuration.DefaultDatatype.aslayer:
            datatype = layer
        elif policy == gdshelpers.configuration.DefaultDatatype.dose_factor:
            datatype = raith_eline_dosefactor_to_datatype(gdshelpers.configuration.dose_factor)

    objs = objs.get_shapely_object() if hasattr(objs, 'get_shapely_object') else objs

    # Convert a list of objects to a true shapely object or collection first
    if type(objs) in [list, tuple]:
        objs = geometric_union(objs)

    # Convert the object or the collection to a list of basic shapely objects
    objs = shapely_collection_to_basic_objs(objs)

    fractured_objs = itertools.chain(*[fracture_intelligently(obj, max_points=max_points,
                                                              max_points_line=max_points_line,
                                                              over_fracture_factor=over_fracture_factor)
                                       for obj in objs if not obj.is_empty])

    exports_objs = list()
    for obj in fractured_objs:
        if type(obj) == shapely.geometry.Point:
            # raise TypeError('Shapely points are not supported to be converted.')
            pass
        elif type(obj) == shapely.geometry.LineString:
            if path_width:
                if library == 'gdscad':
                    exports_objs.append(gdsCAD.core.Path(obj.coords, layer=layer, datatype=datatype, width=path_width,
                                                         pathtype=path_pathtype))
                elif library == 'gdspy':
                    exports_objs.append(gdspy.PolyPath(obj.coords, layer=layer, datatype=datatype, width=path_width,
                                                       ends=path_pathtype))
                elif library == 'oasis':
                    rounded_coords = np.multiply(obj.coords, grid_steps_per_micron).astype(np.int)
                    exports_objs.append(
                        fatamorgana.records.Path((rounded_coords[1:] - rounded_coords[:-1]).tolist(), layer=layer,
                                                 datatype=datatype, half_width=int(path_width * grid_steps_per_micron),
                                                 extension_start=path_pathtype, extension_end=path_pathtype,
                                                 x=rounded_coords[0][0], y=rounded_coords[0][1]))

        elif type(obj) == shapely.geometry.Polygon:
            assert len(obj.interiors) <= 1, 'No polygons with more than one hole allowed, got %i' % len(obj.interiors)
            if library == 'gdscad':
                exports_objs.append(gdsCAD.core.Boundary(obj.exterior.coords, layer=layer, datatype=datatype))
            elif library == 'gdspy':
                exports_objs.append(gdspy.Polygon(obj.exterior.coords, layer=layer, datatype=datatype))
            elif library == 'oasis':
                rounded_coords = np.multiply(obj.exterior.coords, grid_steps_per_micron).astype(np.int)
                exports_objs.append(
                    fatamorgana.records.Polygon((rounded_coords[1:] - rounded_coords[:-1]).tolist(),
                                                layer=layer, datatype=datatype, x=rounded_coords[0][0],
                                                y=rounded_coords[0][1]))

        else:
            raise TypeError('Unhandled type "%s"' % type(obj))
    return exports_objs


def bounds_union(bound_list):
    """
    Calculates the bounding box of all bounding boxes in the given list.
    Each bbox has to be given as (xmin, ymin, xmax, ymax) tuple.

    :param bound_list: List of tuples containing all bounding boxes to be merged
    """
    bounds_array = np.array(bound_list)
    return (np.min(bounds_array[:, 0]), np.min(bounds_array[:, 1]),
            np.max(bounds_array[:, 2]), np.max(bounds_array[:, 3]))


def transform_bounds(bounds, origin, rotation=0, scale=1.):
    """
    Transform a bounds tuple (xmin, ymin, xmax, ymax) by the given offset, rotation and scale.
    """
    bounds = scale * np.array(bounds).reshape(2, 2) + origin
    if rotation != 0:
        center = 0.5 * (bounds[0, :] + bounds[1, :])
        size = np.abs(bounds[1, :] - bounds[0, :])
        c, s = np.cos(rotation), np.sin(rotation)
        rot_matrix = np.array([[c, -s], [s, c]])  # rotation matrix
        new_half_size = 0.5 * np.abs(rot_matrix).dot(size)
        new_center = rot_matrix.dot(center)
        bounds = np.array([new_center - new_half_size, new_center + new_half_size])

    return bounds.flatten()



import numpy as np
from shapely.ops import cascaded_union
from shapely.prepared import prep
from shapely.geometry import Point, LineString, MultiPolygon

from gdshelpers.geometry import geometric_union


def surround_with_holes(geometry, hole_spacing, hole_radius, padding, max_distance):
    """
    Surrounds the given geometry with holes, which are arranged in a square lattice around the structure.
    This can be used for generating vortex traps like presented in https://doi.org/10.1103/PhysRevApplied.11.064053

    :param geometry: The geometry around which the holes are generated
    :param hole_spacing: Spacing between the holes
    :param hole_radius: Radius of the holes
    :param padding: Padding around the geometry
    :param max_distance: Maximum distance of a hole from the geometry
    :return: Shapely object, which describes the holes
    """
    geometry = geometric_union(geometry if isinstance(geometry, (tuple, list)) else (geometry,))
    buffer_around_waveguide = geometry.buffer(max_distance)
    area_for_holes = prep(buffer_around_waveguide.difference(geometry.buffer(hole_radius + padding)))
    area = buffer_around_waveguide.bounds
    points = (Point(x, y) for x in np.arange(area[0], area[2], hole_spacing) for y in
              np.arange(area[1], area[3], hole_spacing))
    return MultiPolygon([point.buffer(hole_radius) for point in points if area_for_holes.contains(point)])


def fill_waveguide_with_holes_in_honeycomb_lattice(waveguide, spacing, padding, hole_radius):
    """
    Fills a given waveguide with holes which are arranged in a honeycomb structure
    This can be used for generating vortex traps like presented in https://doi.org/10.1103/PhysRevApplied.11.064053

    :param waveguide: Waveguide to be filled
    :param spacing: Spacing between the holes
    :param padding: Minimum distance from the edge of the waveguide to the holes
    :param hole_radius: Radius of the holes
    :return: Shapely object, which describes the holes
    """
    center_coordinates = LineString(waveguide.center_coordinates)
    outline = prep(waveguide.get_shapely_outline().buffer(hole_radius).buffer(-padding - 2 * hole_radius))
    area_for_holes = prep(waveguide.get_shapely_object().buffer(hole_radius).buffer(-padding - 2 * hole_radius))
    circles = []

    offset = 0
    new_circles = True
    while new_circles:
        new_circles = False
        for i, pos in enumerate(np.arange(padding, center_coordinates.length - padding, np.sqrt(3 / 4) * spacing)):
            xy = np.array(center_coordinates.interpolate(pos))
            diff = np.array(center_coordinates.interpolate(pos + padding / 2)) - np.array(
                center_coordinates.interpolate(pos - padding / 2))
            d1 = np.array((-diff[1], diff[0])) / np.linalg.norm(diff)
            for direction in [-1, 1]:
                point = Point(xy + direction * (offset + spacing / 2 * (i % 2)) * d1)
                if outline.contains(point):
                    new_circles = True
                if area_for_holes.contains(point):
                    circles.append(point.buffer(hole_radius))
        offset += spacing
    return cascaded_union(circles)


if __name__ == '__main__':
    from gdshelpers.geometry.chip import Cell
    from gdshelpers.parts.waveguide import Waveguide

    wg = Waveguide((0, 0), 0, [3, 3, 3, 3, 3])
    wg.add_straight_segment(5)
    wg.add_bend(np.pi / 2, 20)
    wg.width = 15
    wg.add_straight_segment(4)
    wg.add_bend(-np.pi / 2, 20)

    cell = Cell('vortex_traps')
    cell.add_to_layer(1, wg)
    cell.add_to_layer(2, fill_waveguide_with_holes_in_honeycomb_lattice(wg, 1, .1, .15))
    cell.add_to_layer(2, surround_with_holes(wg.get_shapely_outline(), 3, 1, 1, 15))
    cell.show()
    cell.save()



import unittest
import numpy as np

from gdshelpers.parts.port import Port


class PortTestCase(unittest.TestCase):
    def test_getitem(self):
        origin = (1, 1)
        widths = [2, 1, 2, 1, 2]

        port = Port(origin, 0, widths)

        self.assertAlmostEqual(port[0].origin[1], 1 - 3)
        self.assertAlmostEqual(port[1].origin[1], 1)
        self.assertAlmostEqual(port[2].origin[1], 1 + 3)

        for i, width in enumerate(widths[::2]):
            self.assertAlmostEqual(port[i].width, width)

        self.assertAlmostEqual(port[0:1].origin[1], 1 - 2.5)
        self.assertAlmostEqual(port[0:3].origin[1], 1)
        self.assertAlmostEqual(port[1:3].origin[1], 1 + 1.5)

        for widths in [[2, 1, 2, 1, 2], [2, 1, 2, 1, 2, 1]]:
            port = Port(origin, 0, widths)
            np.testing.assert_almost_equal(port[2].origin, port[-1].origin)
            np.testing.assert_almost_equal(port[1].origin, port[-2].origin)
            np.testing.assert_almost_equal(port[1:3].origin, port[1:-1].origin)
            np.testing.assert_almost_equal(port[1:3].width, port[1:-1].width)
            np.testing.assert_almost_equal(port[2:3].origin, port[-2:-1].origin)
            np.testing.assert_almost_equal(port[2:3].width, port[-2:-1].width)



import os

from gdshelpers.geometry import shapely_adapter


class Shapely3d(object):
    """
    This class saves the values which will be given to blender
    and it also provides a method for output these values
    """

    def __init__(self, poly, extrude_val_min, extrude_val_max, rgb=(255, 255, 255)):
        if type(poly) in (list, tuple):
            self.poly = shapely_adapter.geometric_union(poly)
        self.poly = shapely_adapter.shapely_collection_to_basic_objs(poly)
        self.rgb = rgb
        self.extrude_val_min = extrude_val_min
        self.extrude_val_max = extrude_val_max

    def to_temp(self, scale, index):
        b3d_lines = []
        for obj in self.poly:
            vertices = ["{:f} {:f}\n".format(coord[0], coord[1]) for coord in obj.exterior.coords]
            b3d_lines.extend(["{:d};{:d};{:f};{:f};{:f};{:d};{:d};{:d}\n".format(index, len(vertices),
                                                                                 self.extrude_val_min,
                                                                                 self.extrude_val_max, scale,
                                                                                 self.rgb[0], self.rgb[1],
                                                                                 self.rgb[2])] + vertices)

        return b3d_lines


def save_as_blend(shapely_3d_list, filename, scale=0.1, resolution_x=1980, resolution_y=1080, resolution_percentage=100,
                  render_engine='CYCLES', camera_position_y='above', camera_position_x='right'):
    """
    writes the 'filename.blend' file

    :param shapely_3d_list: a list of the meshes to export
    :param filename: the name of the .blend and .tmp file
    :param scale: size factor
    :param resolution_x: set x resolution for a later render process
    :param resolution_y: set y resolution for a later render process
    :param resolution_percentage: set resolution percentage
    :param render_engine: set the render engine
    :param camera_position_y: decides if the camera is placed 'above' or 'under' the device
    :param camera_position_x: decides if the camera is places 'right' or 'left' of the device
    :return: nothing
    """

    filename = filename[:-6] if filename.endswith('.blend') else filename
    _write_data_and_start_blender(shapely_3d_list, filename, scale, False, True, resolution_x, resolution_y,
                                  resolution_percentage, render_engine, camera_position_y, camera_position_x)


def render_image(shapely_3d_list, filename, scale=0.1, resolution_x=1980, resolution_y=1080, resolution_percentage=100,
                 render_engine='CYCLES', camera_position_y='above', camera_position_x='right'):
    """
    writes the 'filename.png' file

    :param shapely_3d_list: a list of the meshes to export
    :param filename: the name of the .png and .tmp file
    :param scale: size factor
    :param resolution_x: set x resolution for a later render process
    :param resolution_y: set y resolution for a later render process
    :param resolution_percentage: set resolution percentage
    :param render_engine: set the render engine
    :param camera_position_y: decides if the camera is placed 'above' or 'under' the device
    :param camera_position_x: decides if the camera is places 'right' or 'left' of the device
    :return: nothing
    """

    filename = filename[:-4] if filename.endswith('.png') else filename
    _write_data_and_start_blender(shapely_3d_list, filename, scale, True, False,
                                  resolution_x, resolution_y, resolution_percentage,
                                  render_engine, camera_position_y, camera_position_x)


def render_image_and_save_as_blend(shapely_3d_list, filename, scale=0.1, resolution_x=1980, resolution_y=1080,
                                   resolution_percentage=100, render_engine='CYCLES', camera_position_y='above',
                                   camera_position_x='right'):
    """
    writes the 'filename.png' and 'filename.blend' file

    :param shapely_3d_list: a list of the meshes to export
    :param filename: the name of the .png, .blend and .tmp file
    :param scale: size factor
    :param resolution_x: set x resolution for a later render process
    :param resolution_y: set y resolution for a later render process
    :param resolution_percentage: set resolution percentage
    :param render_engine: set the render engine
    :param camera_position_y: decides if the camera is placed 'above' or 'under' the device
    :param camera_position_x: decides if the camera is places 'right' or 'left' of the device
    :return: nothing
    """

    filename = filename[:-4] if filename.endswith('.png') else filename
    _write_data_and_start_blender(shapely_3d_list, filename, scale, True, True, resolution_x, resolution_y,
                                  resolution_percentage, render_engine, camera_position_y, camera_position_x)


def _write_data_and_start_blender(shapely_3d_list, filename, scale, render, save_as_blend_file, resolution_x,
                                  resolution_y, resolution_percentage, render_engine, camera_position_y,
                                  camera_position_x):
    """
    handles the file operation

    :param shapely_3d_list: a list of the meshes to export
    :param filename: the name of the .png, .blend and .tmp file
    :param scale: size factor
    :param resolution_x: set x resolution for a later render process
    :param resolution_y: set y resolution for a later render process
    :param resolution_percentage: set resolution percentage
    :param render_engine: set the render engine
    :param camera_position_y: decides if the camera is placed 'above' or 'under' the device
    :param camera_position_x: decides if the camera is places 'right' or 'left' of the device
    :return: nothing
    """
    with open(filename + ".tmp", "w") as f_out:
        f_out.write("{:b};{:b};{:d};{:d};{:d};{:s};{:s};{:s}".format(render, save_as_blend_file, resolution_x,
                                                                     resolution_y, resolution_percentage, render_engine,
                                                                     camera_position_y, camera_position_x))
        f_out.write('\n')
        for index, s3d in enumerate(shapely_3d_list):
            f_out.writelines([line for line in s3d.to_temp(scale, index)])

    directory = os.path.dirname(os.path.abspath(__file__))

    os.system("blender --background --python {:s}{:s}blender_import.py -- {:s}".format(directory, os.sep, filename))


def _example_i():
    from gdshelpers.parts.splitter import MMI

    mmi1 = MMI((0, 0), 0, 1, 42, 7.7, 2, 2)
    mmi2 = MMI((0, 10), 0, 1, 42, 7.7, 2, 2)
    mmi3 = MMI((0, -10), 0, 1, 42, 7.7, 2, 2)

    shapely_objects = []
    s3d1 = Shapely3d(mmi1.get_shapely_object(), 0.0, 0.7, (0, 0, 0))
    s3d2 = Shapely3d(mmi2.get_shapely_object(), 0.0, 0.7, (0, 255, 0))
    s3d3 = Shapely3d(mmi3.get_shapely_object(), 0.0, 0.7, (255, 0, 0))
    shapely_objects.append(s3d1)
    shapely_objects.append(s3d2)
    shapely_objects.append(s3d3)

    render_image_and_save_as_blend(shapely_objects, "mmi1")


def _example_ii():
    from gdshelpers.parts.splitter import MMI
    from gdshelpers.geometry import geometric_union

    mmi1 = MMI((0, 0), 0, 1, 42, 7.7, 2, 2)
    mmi2 = MMI((0, 10), 0, 1, 42, 7.7, 2, 2)
    mmi3 = MMI((0, -10), 0, 1, 42, 7.7, 2, 2)

    mmi4 = MMI((0, -20), 0, 1, 42, 7.7, 2, 2)
    mmi5 = MMI((0, -30), 0, 1, 42, 7.7, 2, 2)

    shapely_objects = []

    s3d1 = Shapely3d(geometric_union([mmi1.get_shapely_object(), mmi2.get_shapely_object(), mmi3.get_shapely_object()]),
                     0.0, 0.7, (255, 0, 0))

    s3d2 = Shapely3d(geometric_union([mmi4.get_shapely_object(), mmi5.get_shapely_object()]), 0.0, 0.7, (0, 255, 0))

    shapely_objects.append(s3d1)
    shapely_objects.append(s3d2)

    render_image_and_save_as_blend(shapely_objects, "mmi2")


def _example_iii():
    from gdshelpers.parts.waveguide import Waveguide
    from gdshelpers.parts.coupler import GratingCoupler
    from gdshelpers.geometry import geometric_union
    import numpy as np

    coupler1 = GratingCoupler.make_traditional_coupler((250 / 2, 0), 1.3, np.deg2rad(40), 1.13, 0.85, 20,
                                                       taper_length=16, ap_max_ff=0.985, n_ap_gratings=10)
    wave_guide = Waveguide.make_at_port(coupler1.port)
    wave_guide.add_straight_segment(20)
    wave_guide.add_bend(0.5 * np.pi, 40)
    wave_guide.add_straight_segment(250 - 2 * 40)
    wave_guide.add_bend(0.5 * np.pi, 40)
    wave_guide.add_straight_segment(20)
    coupler2 = GratingCoupler.make_traditional_coupler((wave_guide.current_port.origin[0],
                                                        wave_guide.current_port.origin[1]), 1.3, np.deg2rad(40), 1.13,
                                                       0.85, 20, taper_length=16, ap_max_ff=0.985, n_ap_gratings=10)

    shapely_objects = [Shapely3d(geometric_union([coupler1.get_shapely_object(), wave_guide.get_shapely_object(),
                                                  coupler2.get_shapely_object()]), 0.0, 0.7, (255, 255, 255))]

    render_image_and_save_as_blend(shapely_objects, "test_device_under_right", camera_position_y='under',
                                   camera_position_x='right')

    render_image_and_save_as_blend(shapely_objects, "test_device_above_left", camera_position_y='above',
                                   camera_position_x='left')

    render_image_and_save_as_blend(shapely_objects, "test_device_under_left", camera_position_y='under',
                                   camera_position_x='left')

    render_image_and_save_as_blend(shapely_objects, "test_device_above_right", camera_position_y='above',
                                   camera_position_x='right')


if __name__ == '__main__':
    _example_i()
    _example_ii()
    _example_iii()



import numpy as np
import shapely.geometry
from shapely.geometry import Point

from gdshelpers.parts.port import Port
from gdshelpers.helpers import StandardLayers
from gdshelpers.parts.waveguide import Waveguide


class PhotonicCrystalCavity:
    """
    class to implement standard Photonic Crystal Cavity devices with GDS helpers
    see below for examples how to implement cavities with
    1. constant hole diameters/distances
    2. varying hole diameters/distances
    3. tapered waveguide width (set cavity length to - holediameter)
    """

    def __init__(self, origin, angle, width, lengthofcavity, numberofholes=14, holediameters=.5, holedistances=0.5,
                 tapermode=None, finalwidth=None, taperlength=None,
                 samplepoints=1000, holeparams=None, underetching=True, markers=True):
        """
        PhotonicCrystal Cavity

        :param origin: vector, Position of the center of the Photonic Crystal cavity
        :param angle: angle of the roation
        :param width: width of the waveguide
        :param lengthofcavity:  float, distance between the central two points, set to -1*holediameters for
            hole centered cavities
        :param numberofholes:  int, number of holes on each side
        :param holediameters: float or List of the diameters of the holes on each side.
            In case of linear increase use e.g. np.linespace
        :param holedistances: float or List of the distances of the holes on each side.
            In case of linear increase use e.g. np.linespace
        :param tapermode: None for constant widht, 'quadratic' for quadratic tapering
        :param finalwidth: width in the the center of the cavity
        :param taperlength: length over which the cavity will be tapered
        :param samplepoints: number of samplepoints to set the smoothness of the tapering
        """
        if holeparams:
            holetot2 = np.concatenate(
                [np.linspace(holeparams['mindia'], holeparams['maxdia'], holeparams['numholestap']),
                 np.linspace(holeparams['maxdia'], holeparams['maxdia'], holeparams['numholesmir'])])
            distot2 = np.concatenate(
                [np.linspace(holeparams['mindist'], holeparams['maxdist'], holeparams['numholestap']),
                 np.linspace(holeparams['maxdist'], holeparams['maxdist'], (holeparams['numholesmir'] - 1))])
            holediameters = holetot2
            holedistances = distot2
            numberofholes = holeparams['numholestap'] + holeparams['numholesmir']

        self.origin = origin
        self.allholes = list()
        self.width = width
        self.angle = angle
        self._center_portr = Port(origin, angle, self.width)
        self._center_portl = Port(origin, (angle + np.pi), self.width)

        self.widthdiff = finalwidth - width if finalwidth else 0
        self.tapermode = tapermode
        self.lenofcav = lengthofcavity
        if type(holedistances) is float:
            meanholed = holedistances
            holedistances = [holedistances] * (numberofholes - 1)
        else:
            meanholed = np.mean(holedistances)

        if type(holediameters) is float:
            holediameters = [holediameters] * numberofholes

        lenoffset = holediameters[0] + holediameters[-1]

        self.holediameters = holediameters
        self.holedistances = holedistances
        self.numberofholes = len(holediameters)
        self.devlen = lengthofcavity + 2 * ((self.numberofholes - 1) * meanholed) + lenoffset
        self.taperlength = taperlength if taperlength else self.devlen / 2
        self.samplepoints = samplepoints
        self.underetchspace = 10
        self.invertmarker = False
        self.markersize = 20
        self.markerdistancex = 100
        self.markerdistancey = 100

        self.layer_photonic = []
        self.layer_photonic_cavity = []
        self.layer_underetch = []
        self.layer_marker = []
        if len(holediameters) != len(holedistances) + 1:
            print('dimension of holediameters should be 1 more than holedistances!')
            np.append(holediameters, [holediameters[-1]])
        self._generate()
        if underetching:
            self.generate_underetch()
        if markers:
            self.generate_marker()
        if markers == 'inverse':
            self.invertmarker = True
            self.generate_marker()

    def _generate(self):
        if not self.tapermode:
            def width_func(x):
                return self.width

            restlength = 0
        elif self.tapermode == 'quadratic':
            def width_func(x):
                return self.widthdiff * ((1 - x) ** 2) + self.width

            restlength = self.devlen / 2 - self.taperlength

        wgcavl = Waveguide.make_at_port(self._center_portl)
        wgcavl.add_parameterized_path(path=lambda x: (self.taperlength * x, 0),
                                      width=width_func,
                                      sample_points=self.samplepoints, sample_distance=None)
        wgcavl.add_straight_segment(length=restlength)
        self.leftport = wgcavl.current_port
        wgcavr = Waveguide.make_at_port(self._center_portr)
        wgcavr.add_parameterized_path(path=lambda x: (self.taperlength * x, 0),
                                      width=width_func,
                                      sample_points=self.samplepoints, sample_distance=None)
        wgcavr.add_straight_segment(length=restlength)
        self.rightport = wgcavr.current_port
        positionl = [self.origin[0] - (self.lenofcav + self.holediameters[0]) / 2 * np.cos(self.angle),
                     self.origin[1] - (self.lenofcav + self.holediameters[0]) / 2 * np.sin(self.angle)]
        positionr = [self.origin[0] + (self.lenofcav + self.holediameters[0]) / 2 * np.cos(self.angle),
                     self.origin[1] + (self.lenofcav + self.holediameters[0]) / 2 * np.sin(self.angle)]
        wgcavl = wgcavl.get_shapely_object()
        wgcavr = wgcavr.get_shapely_object()
        pointsl = Point(positionl).buffer(distance=self.holediameters[0] / 2)
        wgcavl = wgcavl.difference(pointsl)
        pointsr = Point(positionr).buffer(distance=self.holediameters[0] / 2)
        self.allholes.append(pointsl)
        self.allholes.append(pointsr)
        wgcavr = wgcavr.difference(pointsr)

        for i in range(1, self.numberofholes - 1):
            positionl = [positionl[0] - self.holedistances[i] * np.cos(self.angle),
                         positionl[1] - self.holedistances[i] * np.sin(self.angle)]
            pointsl = Point(positionl).buffer(distance=self.holediameters[i + 1] / 2)
            self.allholes.append(pointsl)
            wgcavl = wgcavl.difference(pointsl)
            positionr = [positionr[0] + self.holedistances[i] * np.cos(self.angle),
                         positionr[1] + self.holedistances[i] * np.sin(self.angle)]
            pointsr = Point(positionr).buffer(distance=self.holediameters[i + 1] / 2)
            self.allholes.append(pointsr)
            wgcavr = wgcavr.difference(pointsr)

        self.layer_photonic_cavity.extend([wgcavl, wgcavr])

    def generate_underetch(self):
        or_r = self.devlen + self.origin[0]  # self.rightport.origin[0]
        or_l = -self.devlen + self.origin[0]  # self.leftport.origin[0]
        underetchbox = shapely.geometry.box(or_l - self.underetchspace,
                                            -(self.width + self.widthdiff + self.underetchspace) + self.origin[1],
                                            or_r + self.underetchspace,
                                            (self.width + self.widthdiff + self.underetchspace) + self.origin[1])

        self.layer_underetch.append((underetchbox))

    def generate_marker(self):
        sign = 1
        if self.invertmarker:
            sign = -1

        markerdistancex = sign * self.markerdistancex / 2
        markerdistancey = sign * self.markerdistancey / 2
        marker_tr = shapely.geometry.box(markerdistancex + self.origin[0] + self.markersize / 2,
                                         markerdistancey + self.origin[1] + self.markersize / 2,
                                         markerdistancex + self.origin[0] - self.markersize / 2,
                                         markerdistancey + self.origin[1] - self.markersize / 2)
        marker_tl = shapely.geometry.box(-markerdistancex + self.origin[0] + self.markersize / 2,
                                         markerdistancey + self.origin[1] + self.markersize / 2,
                                         -markerdistancex + self.origin[0] - self.markersize / 2,
                                         markerdistancey + self.origin[1] - self.markersize / 2)
        marker_b = shapely.geometry.box(self.origin[0] + self.markersize / 2,
                                        -markerdistancey + self.origin[1] + self.markersize / 2,
                                        self.origin[0] - self.markersize / 2,
                                        -markerdistancey + self.origin[1] - self.markersize / 2)

        self.layer_marker.extend([marker_tr, marker_tl, marker_b])

    def get_left_port(self):
        # somehow only works for tapered cavities...
        return self.leftport

    def get_right_port(self):
        # somehow only works for tapered cavities...
        return self.rightport

    def get_holes_list(self):
        # returns a list of all the holes as objects
        return self.allholes

    @classmethod
    def make_at_port(cls, port, **kwargs):
        cavityparameter = dict(kwargs)
        cavityparameter.__delitem__('angle')
        cavityparameter.__delitem__('origin')

        devlen = cavityparameter['lengthofcavity'] + 2 * (
                (cavityparameter['numberofholes'] - 1) * np.mean(cavityparameter['holedistances'])) + (
                     cavityparameter['holediameters']) + (cavityparameter['holediameters'])
        return cls(
            origin=[port.origin[0] + devlen / 2 * np.cos(port.angle), port.origin[1] + devlen / 2 * np.sin(port.angle)],
            angle=port.angle, **cavityparameter)


def _example():
    import gdsCAD
    from gdshelpers.geometry import convert_to_gdscad

    devicename = 'Cavity'
    # cavitiyparameter_const = {'origin': [0, 0], 'angle': 0, 'width': 0.726, 'lengthofcavity': 0.26,
    #                          'numberofholes': 17, 'holediameters': 0.416, 'holedistances': 0.53}
    # cavitiyparameter_lin = {'origin': [100, 500], 'angle': 0, 'width': 0.726, 'lengthofcavity': 0.26,
    #                        'numberofholes': 17,
    #                        'holediameters': np.linspace(0.233, 0.416, 17),
    #                        'holedistances': np.linspace(0.53, 0.605, 16)}
    # taperlength = 15 * 0.414
    # cavitiyparameter_tap_width = {'origin': [-50, -30], 'angle': np.pi, 'width': 1.1, 'numberofholes': 25,
    #                              'taperlength': taperlength,
    #                              'lengthofcavity': -0.414, 'finalwidth': 1.55, 'tapermode': 'quadratic',
    #                              'holediameters': 0.414,
    #                              'holedistances': 0.620}
    holeparams_bg = {'mindia': 0.267, 'maxdia': 0.267, 'mindist': 0.512, 'maxdist': 0.512, 'numholestap': 1,
                     'numholesmir': 10}
    bandgapparameter = {'origin': [0, 0], 'angle': 0,
                        'width': 1.015, 'lengthofcavity': -0.512 / 2,
                        'holeparams': holeparams_bg}

    # pccav1 = PhotonicCrystalCavity(**cavitiyparameter_const)
    bandgap = PhotonicCrystalCavity(**bandgapparameter)

    port1 = Port(origin=[10, 10], angle=np.pi / 2, width=1)
    # cav2 = PhotonicCrystalCavity.make_at_port(port=port1, **cavitiyparameter_tap_width)
    # pccav3 = PhotonicCrystalCavity(**cavitiyparameter_tap_width)
    wg = Waveguide.make_at_port(port1.inverted_direction)
    wg.add_straight_segment(length=5)

    cell = gdsCAD.core.Cell(devicename)
    # cell.add(convert_to_gdscad(pccav1.layer_photonic_cavity, layer=1))
    # cell.add(convert_to_gdscad(pccav2.layer_marker, layer=StandardLayers.lmarklayer))
    # cell.add(convert_to_gdscad(pccav2.layer_underetch, layer=StandardLayers.masklayer1))
    # cell.add(convert_to_gdscad(pccav2.layer_photonic_cavity, layer=StandardLayers.nanolayer))
    # cell.add(convert_to_gdscad(pccav3.layer_marker, layer=StandardLayers.lmarklayer))
    # cell.add(convert_to_gdscad(pccav3.layer_underetch, layer=StandardLayers.masklayer1))
    cell.add(convert_to_gdscad(bandgap.layer_underetch, layer=StandardLayers.masklayer1))
    cell.add(convert_to_gdscad(bandgap.layer_photonic_cavity, layer=StandardLayers.nanolayer))

    # cell.add(convert_to_gdscad(pccav2.layer_photonic_cavity, layer=1))
    # cell.add(convert_to_gdscad(pccav3.get_holes_list()))
    # holes = pccav3.get_holes_list()

    layout = gdsCAD.core.Layout()
    layout.add(cell=cell)
    layout.show()


if __name__ == '__main__':
    _example()



import numpy as np

from gdshelpers.parts.waveguide import Waveguide
from gdshelpers.parts import Port


class StripToSlotModeConverter:
    """
    Generates a strip to slot mode converter as presented by Palmer et. al https://doi.org/10.1109/JPHOT.2013.2239283.
    If the input port width is a scalar und the final width is an array, a strip to slot mode converter is generated.
    On the other hand, if the input port width is an array and the final width is a scaler, a slot to strip mode
    converter is generated.

    """

    def __init__(self, origin, angle, width, taper_length, final_width, pre_taper_length, pre_taper_width):
        """

        :param origin: origin of the mode converter
        :param angle: angle of the mode converter
        :param width: width of the mdoe converter. Scalar if strip to slot, array if slot to strip
        :param taper_length: length of the taper
        :param final_width: final width of the mode converter. Array if strip to slot, scalar if slot to strip
        :param pre_taper_length: length of the pre taper
        :param pre_taper_width: width of the pre taper
        """
        self._taper_length = taper_length
        self._in_port = Port(origin, angle, width)
        self._final_width = final_width
        self._pre_taper_length = pre_taper_length
        self._pre_taper_width = pre_taper_width
        self._waveguide = None

    @classmethod
    def make_at_port(cls, port, taper_length, final_width, pre_taper_length, pre_taper_width):
        """

        :param port: port of the taper (origin, angle, width)
        :param taper_length: length of the taper
        :param final_width: final width of the mode converter. Array if strip to slot, scalar if slot to strip
        :param pre_taper_length: length of the pre taper
        :param pre_taper_width:  width of the pre taper
        :return:
        """
        return cls(port.origin, port.angle, port.width, taper_length, final_width, pre_taper_length,
                   pre_taper_width)

    @property
    def in_port(self):
        """
        Returns the input port of the mode converter

        :return: port
        """
        return self._in_port.copy()

    @property
    def out_port(self):
        """
        Returns the output port of the mode converter

        :return: port
        """
        port = self._in_port.longitudinal_offset(self._taper_length + self._pre_taper_length)
        port.width = self._final_width
        return port

    def get_shapely_object(self):
        """
        Generates the mode converter. If the input port width is a scalar und the final width is an array, a strip to
        slot mode converter is generated. On the other hand, if the input port width is an array and the final width is
        a scaler, a slot to strip mode converter is generated.

        :return: shapely object
        """
        if self._waveguide:
            return self._waveguide.get_shapely_object()

        if np.array(self._in_port.width).size == 1:
            # strip to slot mode converter
            strip_width = self.in_port.width
            slot_width = self._final_width
        else:
            # slot to strip mode converter
            slot_width = self.in_port.width
            strip_width = self._final_width

        def pre_taper_width(t):
            return [
                slot_width[0] * t + self._pre_taper_width * (1 - t),
                slot_width[1] + (slot_width[0] - self._pre_taper_width) * (1 - t),
                strip_width,
                slot_width[0] + slot_width[1]
            ]

        def taper_width(t):
            return [slot_width[0], slot_width[1],
                    slot_width[2] * t + strip_width * (1 - t),
                    (slot_width[0] + slot_width[1]) * (1 - t)
                    ]

        self._waveguide = Waveguide.make_at_port(self._in_port)

        if np.array(self._in_port.width).size == 1:
            # strip to slot mode converter
            self._waveguide.add_parameterized_path(lambda t: [t * self._pre_taper_length, 0], width=pre_taper_width)
            self._waveguide.add_parameterized_path(lambda t: [t * self._taper_length, 0], width=taper_width)
        else:
            # slot to strip mode converter
            self._waveguide.add_parameterized_path(lambda t: [t * self._taper_length, 0],
                                                   width=lambda t: taper_width(1 - t))
            self._waveguide.add_parameterized_path(lambda t: [t * self._pre_taper_length, 0],
                                                   width=lambda t: pre_taper_width(1 - t))

        return self._waveguide.get_shapely_object()


def _example():
    from gdshelpers.geometry.chip import Cell
    from gdshelpers.parts.port import Port
    from gdshelpers.parts.waveguide import Waveguide
    from gdshelpers.parts.mode_converter import StripToSlotModeConverter

    wg_1 = Waveguide.make_at_port(Port(origin=(0, 0), angle=0, width=1.2))  # scalar as width -> strip waveguide
    wg_1.add_straight_segment(5)

    mc_1 = StripToSlotModeConverter.make_at_port(wg_1.current_port, 5, [0.4, 0.2, 0.4], 2,
                                                 0.2)  # array as width -> slot waveguide

    wg_2 = Waveguide.make_at_port(mc_1.out_port)
    wg_2.add_bend(angle=np.pi, radius=5)

    mc_2 = StripToSlotModeConverter.make_at_port(wg_2.current_port, 5, 1, 2, 0.2)  # scalar as width -> strip waveguide

    wg_3 = Waveguide.make_at_port(mc_2.out_port)
    wg_3.add_straight_segment(5)

    cell = Cell('CELL')
    cell.add_to_layer(1, wg_1, mc_1, wg_2, mc_2, wg_3)
    cell.show()


if __name__ == '__main__':
    _example()



from itertools import chain

import numpy as np
import ezdxf


def _add_cell_to_dxf(block, cell, max_points, max_line_points):
    for layer, polygons in cell.get_fractured_layer_dict(max_points, max_line_points).items():
        for polygon in polygons:
            if polygon.interiors:
                raise RuntimeError('DXF only supports polygons without holes')
            block.add_lwpolyline(chain(polygon.exterior.coords, polygon.exterior.coords[0:1]),
                                 dxfattribs={'layer': '{:d}-{:d}'.format(layer, layer)})

    for ref in cell.cells:
        dxfattribs = {}
        if ref['angle'] is not None:
            dxfattribs['rotation'] = np.rad2deg(ref['angle'])
        if ref['magnification'] is not None:
            dxfattribs['xscale'] = dxfattribs['yscale'] = ref['magnification']
        if ref['x_reflection']:
            raise NotImplementedError('X_Reflection for dxf_export is not implemented')
        block.add_blockref(ref['cell'].name, ref['origin'], dxfattribs)


def write_cell_to_dxf_file(outfile, cell, max_points=4000, max_line_points=4000, parallel=False):
    """
    Writes the cell to a dxf-file

    :param outfile: file to write to
    :param cell: cell to export
    :param max_points: maximum number of points for a polygon
    :param max_line_points: maximum number of points for a line
    :param parallel: export if parallelized if true
    :return:
    """
    cells = []
    cell_names = []
    layers_and_datatypes = {}

    dxf = ezdxf.new('R2010')
    msp = dxf.modelspace()
    msp.add_blockref(cell.name, (0, 0))

    def add_cells_to_unique_list(start_cell):
        cells.append(start_cell)
        cell_names.append(start_cell.name)
        for c in start_cell.cells:
            if c not in cells:
                if c['cell'].name in cell_names:
                    raise RuntimeError(
                        'Each cell name must be unique, "{}" is used more than once'.format(c['cell'].name))
                add_cells_to_unique_list(c['cell'])
                for layer in c['cell'].layer_dict.keys():
                    if layer not in layers_and_datatypes:
                        layers_and_datatypes[layer] = {layer}  # second "layer" refers to the datatype
                    elif layer not in layers_and_datatypes[layer]:
                        layers_and_datatypes[layer].add(layer)  # second "layer" refers to the datatype

    add_cells_to_unique_list(cell)

    for layer, datatypes in layers_and_datatypes.items():
        for datatype in datatypes:
            dxf.layers.new(name='{:d}-{:d}'.format(layer, datatype))

    if parallel:
        import warnings
        warnings.warn('Parallel dxf-export is not yet implemented, ignoring parallel flag.')
    for c in cells[::-1]:
        block = dxf.blocks.new(c.name)
        _add_cell_to_dxf(block, c, max_points, max_line_points)
    dxf.write(outfile)


if __name__ == '__main__':
    from gdshelpers.parts.port import Port
    from gdshelpers.parts.waveguide import Waveguide
    from gdshelpers.geometry.chip import Cell

    device_cell = Cell('cell')
    start_port = Port(origin=(10, 0), width=1, angle=0)
    waveguide = Waveguide.make_at_port(start_port)
    for i_bend in range(9):
        waveguide.add_bend(angle=np.pi, radius=60 + i_bend * 40)
    device_cell.add_dlw_taper_at_port('A', 2, waveguide.in_port, 30)
    device_cell.add_dlw_taper_at_port('B', 2, waveguide.current_port, 30)
    device_cell.add_to_layer(1, waveguide)

    sub_cell = Cell('sub_cell')
    sub_cell.add_to_layer(1, waveguide)

    device_cell.add_cell(sub_cell, origin=(10, 10), angle=np.pi / 2)

    with open('dxf_export.dxf', 'w') as file:
        write_cell_to_dxf_file(file, device_cell, parallel=True)



outlayer = 1
""" Layer for photonic devices (Waveguides, Couplers, Splitters...) """

gmarklayer = 2
""" Layer for global markers (Crosses, ...) """

lmarklayer = 3
""" Layer for local markers (Crosses, ...) """

wflayer = 4
""" Layer for write fields. (Used just for checking, not used in lithography) """

regionlayer = 5
""" Layer for region definition. (For defining working areas = multiple of writefield_size) """

framelayer = 6
""" Layer for frames. (For overall pattern and/or arrays of devices) """

patnamelayer = 7
""" Layer for pattern names. (For example NG01 CalibCoupler) """

devnamelayer = 8
""" Layer for device names. (For example A1, F0, etc..) """

parnamelayer1 = 9
""" Layer for local parameters. (Period of couplers, gaps, ...  Used for first lithography step) """

parnamelayer2 = 10
""" Layer for local parameters. (Period of couplers, gaps, ...  Used for second lithography step) """

padlayer = 11
""" Metal contact pad layer. (For probe contacts or wire bonding) """

winglayer = 12
""" Wing layer. (For connection between nanowires and pads) """

nanolayer = 13
""" Nano features layer. (SSPD wires, high resolution elements...) """

masklayer1 = 14
""" Mask window layer 1. (opening windows for subsequent etching e.g. freestanding devices) """

masklayer2 = 15
""" Mask window layer 2. (opening windows for subsequent etching e.g. freestanding devices) """

gplayers = tuple(range(16, 70))
""" List of unassigned layers for general purpose use. """



import itertools
import json
import numpy as np
from shapely.affinity import translate, rotate
from shapely.geometry import box

from gdshelpers.geometry.shapely_adapter import convert_to_layout_objs, bounds_union, transform_bounds
from gdshelpers.export.gdsii_export import write_cell_to_gdsii_file
from gdshelpers.geometry import geometric_union
import gdshelpers.helpers.layers as std_layers


class Cell:
    def __init__(self, name):
        """
        Creates a new Cell named `name` at `origin`

        :param name: Name of the cell, needs to be unique
        """
        self.name = name
        self.cells = []
        self.layer_dict = {}
        self.dlw_data = {}
        self.desc = {'dlw': self.dlw_data, 'desc': {}, 'ebl': []}
        self.cell_gdspy = None
        self.cell_gdscad = None
        self.cell_oasis = None
        self._bounds = None
        # Only contains the bounds of items in `layer_dict`.
        # Children cells have to be queried each time, since there is no way of knowing when they got changed.
        # self._bounds is None if the bounds need to be recalculated or if the cell is empty (in which case
        # recalculating them is cheap and we don't need to cache them)

    @property
    def bounds(self):
        """
        The outer bounding box of the cell. Returns `None` if it is empty.
        """
        return self.get_bounds(layers=None)

    def get_bounds(self, layers=None):
        """
        Calculates and returns the envelope for the given layers.
        Returns `None` if it is empty.
        """
        # Go through all layers
        bounds = []
        if layers is None and self._bounds is not None:
            bounds += [self._bounds]
        else:
            for layer in layers or self.layer_dict.keys():
                for geo in self.layer_dict.get(layer, []):
                    geo_bounds = geo.get_shapely_object().bounds if hasattr(geo, 'get_shapely_object') else geo.bounds
                    if geo_bounds is not ():  # Some shapely geometries (collections) can return empty bounds
                        bounds.append(geo_bounds)

            # Cache envelope if we have the global envelope
            if layers is None:
                self._bounds = bounds_union(bounds) if len(bounds) > 0 else None

        # Merge envelopes of children cells
        for cell in self.cells:
            cell_bounds = cell['cell'].get_bounds(layers)
            if cell_bounds is not None:
                bounds.append(transform_bounds(cell_bounds, cell['origin'], rotation=cell['angle'] or 0))

        return bounds_union(bounds) if len(bounds) > 0 else None

    @property
    def size(self):
        """
        Returns the size of the cell
        """
        bounds = self.bounds
        if bounds is None:
            return 0, 0
        else:
            return bounds[2] - bounds[0], bounds[3] - bounds[1]

    def add_to_layer(self, layer, *geometry):
        """
        Adds a shapely geometry to a the layer

        :param layer: id of the layer
        :param geometry: shapely geometry
        """
        self._bounds = None
        if layer not in self.layer_dict:
            self.layer_dict[layer] = []
        self.layer_dict[layer] += geometry

    def add_dlw_data(self, dlw_type, dlw_id, data):
        """
        Adds data for 3D-hybrid-integration to the Cell. This is usually only done by using the Device

        :param dlw_type: type of the represented object
        :param dlw_id: id of the represented object
        :param data: data of the object
        """
        if dlw_type not in self.dlw_data:
            self.dlw_data[dlw_type] = {}
        if dlw_id in self.dlw_data[dlw_type]:
            raise ValueError('ID "{:s}" already used'.format(dlw_id))
        self.dlw_data[dlw_type][dlw_id] = data

    def add_cell(self, cell, origin=(0, 0), angle=None, columns=1, rows=1, spacing=None):
        """
        Adds a Cell to this cell

        :param cell: Cell to add
        :param origin: position where to add the cell
        :param angle: defines the rotation of the cell
        :param columns: Number of columns
        :param rows: Number of rows
        :param spacing: Spacing between the cells, should be an array in the form [x_spacing, y_spacing]
        """
        if cell.name in [cell_dict['cell'].name for cell_dict in self.cells]:
            import warnings
            warnings.warn(
                'Cell name "{cell_name:s}" added multiple times to {self_name:s}.'
                ' Can be problematic for desc/dlw-files'.format(cell_name=cell.name, self_name=self.name))
        self.cells.append(
            dict(cell=cell, origin=origin, angle=angle, magnification=None, x_reflection=False, columns=columns,
                 rows=rows, spacing=spacing))

    def add_region_layer(self, region_layer=std_layers.regionlayer, layers=None):
        """
        Generate a region layer around all objects on `layers` and place it on layer `region_layer`.
        If `layers` is None, all layers are used.
        """
        self.add_to_layer(region_layer, box(*self.get_bounds(layers)))

    def add_frame(self, padding=30., line_width=1., frame_layer=std_layers.framelayer, bounds=None):
        """
        Generates a rectangular frame around the contents of the cell.

        :param padding: Add a padding of the given value around the contents of the cell
        :param line_width: Width of the frame line
        :param frame_layer: Layer to put the frame on.
        :param bounds: Optionally, an explicit extent in the form (min_x, min_y, max_x, max_y) can be passed to
            the function. If `None` (default), the current extent of the cell will be chosen.
        """
        padding = padding + line_width
        bounds = bounds or self.bounds

        frame = box(bounds[0] - padding, bounds[1] - padding, bounds[2] + padding, bounds[3] + padding)
        frame = frame.difference(frame.buffer(-line_width))
        self.add_to_layer(frame_layer, frame)

    def add_ebl_marker(self, layer, marker):
        """
        Adds an Marker to the layout

        :param layer: layer on which the marker should be positioned
        :param marker: marker, that should be added (from gdshelpers.parts.markers)
        """
        self.add_to_layer(layer, marker)
        self.desc['ebl'].append(list(marker.origin))

    def add_ebl_frame(self, layer, frame_generator, bounds=None, **kwargs):
        """
        Adds global markers to the layout

        :param layer:  layer on which the markers should be positioned
        :param frame_generator: either a method, which returns a list of the markers, which should be added or the name
            of a generator from the gdshelpers.geometry.ebl_frame_generators package
        :param bounds: Optionally the bounds to use can be provided in the form (min_x, min_y, max_x, max_y). If None,
            the standard cell bounds will be used.
        :param kwargs: Parameters which are directly passed to the frame generator (other than the bounds parameter)
        """
        from gdshelpers.geometry import ebl_frame_generators
        frame_generator = frame_generator if callable(frame_generator) else getattr(ebl_frame_generators,
                                                                                    frame_generator)
        bounds = bounds or self.bounds

        for marker in frame_generator(bounds, **kwargs):
            self.add_ebl_marker(layer, marker)

    def add_to_desc(self, key, data):
        """
        Adds data to the .desc-file can be used for any data related to the cells, e.g. the swept parameters/...

        :param key: name of the entry
        :param data: data which are added to the .desc-file, the data have to be serializable by json
        """
        self.desc['desc'][key] = data

    def get_dlw_data(self):
        dlw_data = self.dlw_data.copy()
        for sub_cell in self.cells:
            cell, origin = sub_cell['cell'], sub_cell['origin']

            for dlw_type, dlw_type_data in cell.get_dlw_data().items():
                for dlw_id, data in dlw_type_data.items():
                    data = data.copy()
                    data['origin'] = (np.array(origin) + data['origin']).tolist()
                    if dlw_type not in dlw_data:
                        dlw_data[dlw_type] = {}
                    dlw_data[dlw_type][cell.name + '.' + dlw_id] = data

        return dlw_data

    def get_desc(self):
        desc = self.desc.copy()
        desc['cells'] = {cell['cell'].name: dict(offset=tuple(cell['origin']), angle=cell['angle'] or 0,
                                                 **cell['cell'].get_desc()) for cell in self.cells}
        return desc

    def get_fractured_layer_dict(self, max_points=4000, max_line_points=4000):
        from gdshelpers.geometry.shapely_adapter import shapely_collection_to_basic_objs, fracture_intelligently
        fractured_layer_dict = {}
        for layer, geometries in self.layer_dict.items():
            fractured_geometries = []
            for geometry in geometries:
                geometry = geometry.get_shapely_object() if hasattr(geometry, 'get_shapely_object') else geometry
                if type(geometry) in [list, tuple]:
                    geometry = geometric_union(geometry)
                geometry = shapely_collection_to_basic_objs(geometry)
                geometry = itertools.chain(
                    *[fracture_intelligently(geo, max_points, max_line_points) for geo in geometry if not geo.is_empty])
                fractured_geometries.append(geometry)
            fractured_layer_dict[layer] = itertools.chain(*fractured_geometries)
        return fractured_layer_dict

    def get_gdspy_cell(self, executor=None):
        import gdspy
        if self.cell_gdspy is None:
            self.cell_gdspy = gdspy.Cell(self.name)
            for sub_cell in self.cells:
                angle = np.rad2deg(sub_cell['angle']) if sub_cell['angle'] is not None else None
                if sub_cell['columns'] == 1 and sub_cell['rows'] == 1 and not sub_cell['spacing']:
                    self.cell_gdspy.add(
                        gdspy.CellReference(sub_cell['cell'].get_gdspy_cell(executor), origin=sub_cell['origin'],
                                            rotation=angle, magnification=sub_cell['magnification'],
                                            x_reflection=sub_cell['x_reflection']))
                else:
                    self.cell_gdspy.add(
                        gdspy.CellArray(sub_cell['cell'].get_gdspy_cell(executor), origin=sub_cell['origin'],
                                        rotation=angle, magnification=sub_cell['magnification'],
                                        x_reflection=sub_cell['x_reflection'], columns=sub_cell['columns'],
                                        rows=sub_cell['rows'], spacing=sub_cell['spacing']))
            for layer, geometries in self.layer_dict.items():
                for geometry in geometries:
                    if executor:
                        executor.submit(convert_to_layout_objs, geometry, layer, library='gdspy') \
                            .add_done_callback(lambda future: self.cell_gdspy.add(future.result()))
                    else:
                        self.cell_gdspy.add(convert_to_layout_objs(geometry, layer, library='gdspy'))
        return self.cell_gdspy

    def get_gdscad_cell(self, executor=None):
        import gdsCAD
        if self.cell_gdscad is None:
            self.cell_gdscad = gdsCAD.core.Cell(self.name)
            for sub_cell in self.cells:
                angle = np.rad2deg(sub_cell['angle']) if sub_cell['angle'] is not None else None
                if sub_cell['columns'] == 1 and sub_cell['rows'] == 1 and not sub_cell['spacing']:
                    self.cell_gdscad.add(
                        gdsCAD.core.CellReference(sub_cell['cell'].get_gdscad_cell(executor), origin=sub_cell['origin'],
                                                  rotation=angle, magnification=sub_cell['magnification'],
                                                  x_reflection=sub_cell['x_reflection']))
                else:
                    self.cell_gdscad.add(
                        gdsCAD.core.CellArray(sub_cell['cell'].get_gdscad_cell(executor), origin=sub_cell['origin'],
                                              rotation=angle, magnification=sub_cell['magnification'],
                                              x_reflection=sub_cell['x_reflection'], cols=sub_cell['columns'],
                                              rows=sub_cell['rows'], spacing=sub_cell['spacing']))
            for layer, geometries in self.layer_dict.items():
                for geometry in geometries:
                    if executor:
                        executor.submit(convert_to_layout_objs, geometry, layer, library='gdscad') \
                            .add_done_callback(lambda future: self.cell_gdscad.add(future.result()))
                    else:
                        self.cell_gdscad.add(convert_to_layout_objs(geometry, layer, library='gdscad'))
        return self.cell_gdscad

    def get_oasis_cells(self, grid_steps_per_micron=1000, executor=None):
        import fatamorgana
        import fatamorgana.records
        if self.cell_oasis is None:
            self.cell_oasis = fatamorgana.Cell(fatamorgana.NString(self.name))
            for sub_cell in self.cells:
                x, y = sub_cell['origin']
                x, y = round(x * grid_steps_per_micron), round(y * grid_steps_per_micron)
                angle = np.rad2deg(sub_cell['angle']) if sub_cell['angle'] is not None else None
                repetition = None
                if not (sub_cell['columns'] == 1 and sub_cell['rows'] == 1 and not sub_cell['spacing']):
                    repetition = fatamorgana.basic.GridRepetition(
                        [round(sub_cell['spacing'][0] * grid_steps_per_micron), 0],
                        sub_cell['columns'],
                        [0, round(sub_cell['spacing'][1] * grid_steps_per_micron)],
                        sub_cell['rows'])
                self.cell_oasis.placements.append(
                    fatamorgana.records.Placement(False, name=fatamorgana.NString(sub_cell['cell'].name), x=x, y=y,
                                                  angle=angle, repetition=repetition))
            for layer, geometries in self.layer_dict.items():
                for geometry in geometries:
                    if executor:
                        executor.submit(convert_to_layout_objs, geometry, layer, library='oasis',
                                        grid_steps_per_micron=grid_steps_per_micron, max_points=np.inf,
                                        max_points_line=np.inf) \
                            .add_done_callback(lambda future: self.cell_oasis.geometry.extend(future.result()))
                    else:
                        self.cell_oasis.geometry.extend(
                            convert_to_layout_objs(geometry, layer, library='oasis',
                                                   grid_steps_per_micron=grid_steps_per_micron,
                                                   max_points=np.inf, max_points_line=np.inf))
        return [self.cell_oasis] + [oasis_cell for sub_cell in self.cells for oasis_cell in
                                    sub_cell['cell'].get_oasis_cells(grid_steps_per_micron, executor)]

    def get_gdspy_lib(self):
        import gdspy
        self.get_gdspy_cell()
        return gdspy.current_library

    def start_viewer(self):
        import gdspy
        gdspy.LayoutViewer(library=self.get_gdspy_lib(), depth=10)

    def save(self, name=None, library=None, grid_steps_per_micron=1000, parallel=False):
        """
        Exports the layout and creates an DLW-file, if DLW-features are used.

        :param name: The filename of the saved file. The ending of the filename defines the format.
            Currently .gds, .oasis and .dxf are supported.
        :param library: Name of the used library.
            Should stay `None` in order to select the library depending on the file-ending.
            The use of this parameter is deprecated and this parameter will be removed in a future release.
        :param grid_steps_per_micron: Defines the resolution
        :param parallel: Defines if parallelization is used (only supported in Python 3).
            Standard value will be changed to True in a future version.
            Deactivating can be useful for debugging reasons.
        """

        if library is not None:
            import warnings
            warnings.warn('The use of the `library` parameter is deprecated. '
                          'Instead, define the format by the file ending of the filename.', DeprecationWarning)

        if not name:
            name = self.name
        elif name.endswith('.gds'):
            name = name[:-4]
            library = library or 'gdshelpers'
        elif name.endswith('.oasis'):
            name = name[:-6]
            library = library or 'fatamorgana'
        elif name.endswith('.dxf'):
            name = name[:-4]
            library = library or 'ezwriter'

        library = library or 'gdshelpers'

        if library == 'gdshelpers':
            from tempfile import NamedTemporaryFile
            import shutil

            with NamedTemporaryFile('wb', delete=False) as tmp:
                write_cell_to_gdsii_file(tmp, self, grid_steps_per_unit=grid_steps_per_micron, parallel=parallel)
            shutil.move(tmp.name, name + '.gds')

        elif library == 'gdspy':
            import gdspy

            if parallel:
                from concurrent.futures import ProcessPoolExecutor
                with ProcessPoolExecutor() as pool:
                    self.get_gdspy_cell(pool)
            else:
                self.get_gdspy_cell()

            self.get_gdspy_lib().precision = self.get_gdspy_lib().unit / grid_steps_per_micron
            gdspy_cells = self.get_gdspy_lib().cell_dict.values()
            if parallel:
                from concurrent.futures import ProcessPoolExecutor
                with ProcessPoolExecutor() as pool:
                    binary_cells = pool.map(gdspy.Cell.to_gds, gdspy_cells, [grid_steps_per_micron] * len(gdspy_cells))
            else:
                binary_cells = map(gdspy.Cell.to_gds, gdspy_cells, [grid_steps_per_micron] * len(gdspy_cells))

            self.get_gdspy_lib().write_gds(name + '.gds', cells=[], binary_cells=binary_cells)
        elif library == 'gdscad':
            import gdsCAD
            layout = gdsCAD.core.Layout(precision=1e-6 / grid_steps_per_micron)
            if parallel:
                from concurrent.futures import ProcessPoolExecutor
                with ProcessPoolExecutor() as pool:
                    layout.add(self.get_gdscad_cell(pool))
            else:
                layout.add(self.get_gdscad_cell())
            layout.save(name + '.gds')
        elif library == 'fatamorgana':
            import fatamorgana
            layout = fatamorgana.OasisLayout(grid_steps_per_micron)

            if parallel:
                from concurrent.futures import ProcessPoolExecutor
                with ProcessPoolExecutor() as pool:
                    cells = self.get_oasis_cells(grid_steps_per_micron, pool)
            else:
                cells = self.get_oasis_cells(grid_steps_per_micron)

            layout.cells = [cells[0]] + list(set(cells[1:]))

            # noinspection PyUnresolvedReferences
            def replace_names_by_ids(oasis_layout):
                name_id = {}
                for cell_id, cell in enumerate(oasis_layout.cells):
                    if cell.name.string in name_id:
                        raise RuntimeError(
                            'Each cell name should be unique, name "' + cell.name.string + '" is used multiple times')
                    name_id[cell.name.string] = cell_id
                    cell.name = cell_id
                for cell in oasis_layout.cells:
                    for placement in cell.placements:
                        placement.name = name_id[placement.name.string]

                oasis_layout.cellnames = {v: k for k, v in name_id.items()}

            # improves performance for reading oasis file and workaround for fatamorgana-bug
            replace_names_by_ids(layout)

            with open(name + '.oas', 'wb') as f:
                layout.write(f)
        elif library == 'ezdxf':
            from gdshelpers.export.dxf_export import write_cell_to_dxf_file
            with open(name + '.gds', 'wb') as f:
                write_cell_to_dxf_file(f, self, grid_steps_per_micron, parallel=parallel)
        else:
            raise ValueError('library must be either "gdscad", "gdspy" or "fatamorgana"')

        dlw_data = self.get_dlw_data()
        if dlw_data:
            with open(name + '.dlw', 'w') as f:
                json.dump(dlw_data, f, indent=True)

    def save_desc(self, filename):
        """
        Saves a description file for the layout. The file format is not final yet and might change in a future release.

        :param filename: name of the file the description data will be written to
        """
        if not filename.endswith('.desc'):
            filename += '.desc'
        with open(filename + '.desc', 'w') as f:
            json.dump(self.get_desc(), f, indent=True)

    def get_reduced_layer(self, layer):
        """
        Returns a single shapely object containing the structures on a certain layer from this cell and all added cells.

        :param layer: the layer whose structures will be returned
        :return: a single shapely-geometry
        """

        def translate_and_rotate(geometry, offset, angle):
            if not geometry:
                return geometry
            return translate(rotate(geometry, angle if angle else 0, use_radians=True, origin=(0, 0)), *offset)

        return geometric_union(
            (self.layer_dict[layer] if layer in self.layer_dict else []) +
            [translate_and_rotate(cell['cell'].get_reduced_layer(layer), cell['origin'], cell['angle'])
             for cell in self.cells])

    def export_mesh(self, filename, layer_defs):
        """
        Saves the current geometry as a mesh-file.

        :param filename: Name of the file which will be created. The file ending determines the format.
        :param layer_defs: Definition of the layers, should be a list like [(layer,(z_min,z_max)),...]
        """
        from functools import reduce
        from trimesh.primitives import Extrusion
        from trimesh.transformations import translation_matrix

        reduce(lambda a, b: a + b, (Extrusion(polygon=geometry, height=min_max[1] - min_max[0],
                                              transform=translation_matrix((0, 0, min_max[0])))
                                    for layer, min_max in layer_defs.items()
                                    for geometry in (lambda x: x if hasattr(x, '__iter__') else [x, ])(
            self.get_reduced_layer(layer)))).export(filename)

    def get_patches(self, origin=(0, 0), angle_sum=0, angle=0, layers=None):
        from descartes import PolygonPatch

        def rotate_pos(pos, rotation_angle):
            if rotation_angle is None:
                return pos
            c, s = np.cos(rotation_angle), np.sin(rotation_angle)
            result = np.array([[c, -s], [s, c]]).dot(pos)
            return result

        own_patches = []
        for layer, geometry in self.layer_dict.items():
            if layers is not None and layer not in layers:
                continue
            geometry = geometric_union(geometry)
            if geometry.is_empty:
                continue
            geometry = translate(rotate(geometry, angle_sum, use_radians=True, origin=(0, 0)), *origin)
            own_patches.append(
                PolygonPatch(geometry, color=['red', 'green', 'blue', 'teal', 'pink'][(layer - 1) % 5], linewidth=0))

        sub_cells_patches = [p for cell_dict in self.cells for p in
                             cell_dict['cell'].get_patches(
                                 np.array(origin) + rotate_pos(cell_dict['origin'], angle),
                                 angle_sum=angle_sum + (cell_dict['angle'] or 0), angle=cell_dict['angle'],
                                 layers=layers)]

        return own_patches + sub_cells_patches

    def save_image(self, filename, layers=None, antialiased=True, resolution=1., ylim=(None, None), xlim=(None, None)):
        """
           Save cell object as an image.

           You can either use a rasterized file format such as png but also formats such as SVG or PDF.

           :param filename: Name of the image file.
           :param resolution: Rasterization resolution in GDSII units.
           :param antialiased: Whether to use a anti-aliasing or not.
           :param ylim: Tuple of (min_x, max_x) to export.
           :param xlim: Tuple of (min_y, max_y) to export.
           """
        import matplotlib.pyplot as plt

        # For vector graphics, map 1um to {resolution} mm instead of inch.
        is_vector = filename.split('.')[-1] in ('svg', 'svgz', 'eps', 'ps', 'emf', 'pdf')
        scale = 5 / 127. if is_vector else 1.

        fig, ax = plt.subplots()
        for patch in self.get_patches(layers=layers):
            patch.set_antialiased(antialiased)
            ax.add_patch(patch)

        # Autoscale, then change the axis limits and read back what is actually displayed
        ax.autoscale(True, tight=True)
        ax.set_xlim(*xlim)
        ax.set_ylim(*ylim)
        actual_ylim, actual_xlim = ax.get_ylim(), ax.get_xlim()
        fig.set_size_inches(np.asarray((actual_xlim[1] - actual_xlim[0], actual_ylim[1] - actual_ylim[0])) * scale)

        ax.set_aspect(1)
        ax.axis('off')

        fig.set_dpi(1 / resolution)
        plt.savefig(filename, transparent=True, bbox_inches='tight', dpi=1 / resolution)
        plt.close()

    def show(self, layers=None, padding=5):
        """
        Shows the current cell

        :param layers: List of the layers to be shown, passing None shows all layers
        :param padding: padding around the structure
        """
        import matplotlib.pyplot as plt

        fig, ax = plt.subplots()
        for patch in self.get_patches(layers=layers):
            ax.add_patch(patch)

        bounds = self.get_bounds(layers)
        ax.set_xlim(bounds[0] - padding, bounds[2] + padding)
        ax.set_ylim(bounds[1] - padding, bounds[3] + padding)
        ax.set_aspect(1)
        fig.show()

    def add_dlw_marker(self, label, layer, origin):
        """
        Adds a marker for 3D-hybrid integration

        :param label: Name of the marker, needs to be unique within the device
        :param layer: Layer at which the marker and markers should be written
        :param origin: Position of the marker
        """
        from gdshelpers.parts.marker import DLWMarker
        from gdshelpers.parts.text import Text

        self.add_to_layer(layer, DLWMarker(origin))
        self.add_to_layer(std_layers.parnamelayer1, Text(origin, 2, label, alignment='center-center'))

        self.add_dlw_data('marker', label, {'origin': origin.tolist()})

    def add_dlw_taper_at_port(self, label, layer, port, taper_length, tip_width=.01, with_markers=True):
        """
        Adds a taper for 3D-hybrid-integration at a certain port

        :param label: Name of the port, needs to be unique within the device
        :param layer: Layer at which the taper and markers should be written
        :param port: Port to which the taper should be attached
        :param taper_length: length of the taper
        :param tip_width: final width of the tip
        :param with_markers: for recognizing the taper markers near to the taper are necessary.
            In certain designs the standard positions are not appropriate and
            can therefore be disabled and manually added
        """
        from gdshelpers.parts.text import Text

        taper_port = port.longitudinal_offset(taper_length)
        if taper_length > 0:
            from gdshelpers.parts.waveguide import Waveguide
            wg = Waveguide.make_at_port(port)
            wg.add_straight_segment(taper_length, final_width=tip_width)
            self.add_to_layer(layer, wg.get_shapely_object())
        self.add_to_layer(std_layers.parnamelayer1, Text(taper_port.origin, 2, label, alignment='center-center'))

        self.add_dlw_data('taper', str(label), {'origin': taper_port.origin.tolist(), 'angle': port.angle,
                                                'starting_width': port.width, 'taper_length': taper_length})
        if with_markers:
            for i, (v, l) in enumerate(itertools.product((-20, 20), (taper_length, 0))):
                self.add_dlw_marker(str(label) + '-' + str(i), layer,
                                    port.parallel_offset(v).longitudinal_offset(l).origin)


if __name__ == '__main__':
    from gdshelpers.parts.port import Port
    from gdshelpers.parts.waveguide import Waveguide

    # Create a cell-like object that offers a save output command '.save' which creates the .gds or .oas file by using
    # gdsCAD,gdspy or fatamorgana
    device_cell = Cell('my_cell')
    # Create a port to connect waveguide structures to
    start_port = Port(origin=(0, 0), width=1, angle=0)
    waveguide = Waveguide.make_at_port(start_port)
    for i_bend in range(9):
        waveguide.add_bend(angle=np.pi, radius=60 + i_bend * 40)
    # Add direct laser writing taper and alignment marker for postprocessing with a dlw printer to the cell-like object.
    # The cell dlw files will be saved with the cell.
    device_cell.add_dlw_taper_at_port('A0', 1, start_port.inverted_direction, 30)
    device_cell.add_dlw_taper_at_port('A1', 1, waveguide.current_port, 30)
    device_cell.add_to_layer(1, waveguide)
    device_cell.show()
    device_cell.save_image('chip.pdf')
    # Creates the output file by using gdspy, gdscad or fatamorgana. To use the implemented parallel processing, set
    # parallel=True.
    device_cell.save(name='my_design', parallel=True)
    device_cell.export_mesh('my_design.stl', layer_defs={1: (0, 1)})

    array_cell = Cell('Array')
    array_cell.add_cell(device_cell, rows=2, columns=2, spacing=(1000, 1000))
    array_cell.save()



from __future__ import division

from math import sqrt
import numpy as np
from shapely.affinity import rotate, translate, scale
from shapely.geometry import Polygon, box
from shapely.ops import cascaded_union


class KITLogo(object):
    """
    A simplified logo of the Karlsruhe Institute of Technology (KIT).

    :param origin: Tuple specifying the lower left corner of the logo.
    :param height: Height of the logo.
    :param min_radius_fraction: To avoid steep angles, the rays are cut at the minimum radius fraction.
    """

    def __init__(self, origin, height, min_radius_fraction=0.05):
        self.origin = origin
        self.height = height
        self.min_radius_fraction = min_radius_fraction

    def get_shapely_object(self):
        # Start with the "rays"
        # noinspection PyTypeChecker
        ray_angles = np.linspace(0, np.pi / 2, 8) + np.pi / 2
        outer_ray_positions = (np.array([np.cos(ray_angles), np.sin(ray_angles)]) * self.height).T + (self.height, 0)
        inner_ray_positions = (np.array([np.cos(ray_angles), np.sin(ray_angles)])
                               * self.height * self.min_radius_fraction).T + (self.height, 0)

        polygons = list()
        for outer, inner in zip(outer_ray_positions.reshape(-1, 2, 2), inner_ray_positions.reshape(-1, 2, 2)):
            polygons.append(Polygon([outer[0], outer[1], inner[1], inner[0]]))
            pass

        # Draw the letters
        d = self.height * 0.077

        k_upper_branch = rotate(box(0, -d, sqrt(2) * self.height / 2 + sqrt(2) * d, d), 45, origin=(0, 0))
        k_lower_branch = scale(k_upper_branch, yfact=-1., origin=(0, 0))
        k_uncut = k_upper_branch.union(k_lower_branch)
        k_unscaled = k_uncut.intersection(box(0, -self.height / 2., self.height / 2. + sqrt(2) * d, self.height / 2.))
        k = scale(k_unscaled, 0.8, origin=(0, 0))
        polygons.append(translate(k, self.height * 1.05, self.height / 2.))

        i = box(0, 0, 2 * d, self.height)
        polygons.append(translate(i, self.height * 1.6))

        t_overlap = 2
        t = box(-d, 0, d, self.height).union(
            box(-d * (1 + t_overlap), self.height - 2 * d, d * (1 + t_overlap), self.height))
        polygons.append(translate(t, self.height * 2.05))

        logo = cascaded_union(polygons)

        return translate(logo, *self.origin)


class WWULogo(object):
    """
    WWU Logo with WWU written next to it

    :param origin: Tuple specifying the lower left corner of the logo.
    :param height: Height of the logo.
    :param text: 0 no text, 1 text right of logo, 2 text below logo
    """

    def __init__(self, origin, height, text):
        self.origin = origin
        self.height = height
        self.text = text

    def get_shapely_object(self):
        # define WWU Logo measures.
        w = (120 + 96 + 120)
        h = 16 + 59 + 8 + 26 + 16 + 14 + 16 + 23 + 8 + 8 + 16

        h1 = 16  # height of box1
        s1 = 59  # vertical spacing between box1 & 2

        h2 = 8
        s2 = 26

        h3 = 16
        w3 = 120
        s3 = 14

        h4 = 16
        w4 = 51
        s4 = 23

        h5 = 8
        w5 = 30
        s5 = 8

        h6 = 16
        w6 = 8

        # calculate scaling factor to match given height
        # M = self.height/h  # scaling factor

        # each Box's lower left corner xi,yi
        x1, y1 = 0, 0
        x2, y2 = x1, y1 + (h1 + s1)
        x3l, y3 = x2, y2 + (h2 + s2)
        x3r = x2 + 120 + 96
        x4, y4 = x1 + (w - w4) / 2, y3 + (h3 + s3)
        x5, y5 = x1 + (w - w5) / 2, y4 + (h4 + s4)
        x6, y6 = x1 + (w - w6) / 2, y5 + (h5 + s5)

        # define boxes
        boxes = [
            box(x1, y1, x1 + w, y1 + h1),  # box1
            box(x2, y2, x2 + w, y2 + h2),  # ...
            box(x3l, y3, x3l + w3, y3 + h3),
            box(x3r, y3, x3r + w3, y3 + h3),
            box(x4, y4, x4 + w4, y4 + h4),
            box(x5, y5, x5 + w5, y5 + h5),
            box(x6, y6, x6 + w6, y6 + h6)  # box6
        ]

        logo_unscaled = cascaded_union(boxes)

        # write WWU
        # create W
        x_w = [-40.58, -13.7, 1.05, 16.86, 40.05, 68.51, 45.06, 29.78, 13.17, -11.33, -28.19, -43.21, -66.93]
        y_w = [0, 0, 78.26, 0, 0, 114, 114, 33.46, 114, 114, 33.46, 114, 114]
        w_coords = zip(x_w, y_w)
        w_shape = Polygon(w_coords)

        # create U
        # ellipse w=82=2a, h=58=2b
        # parametrisierung x=a*cos(t), y=b*sin(t), t=0..2pi
        num = 20
        a = 41
        b = 29

        t = np.linspace(np.pi, 2 * np.pi, num)
        xb = a * np.cos(t)  # bottom coordinates
        yb = b * np.sin(t)

        # right box
        xr = [max(xb), max(xb) - 25]
        yr = [114 + min(yb), 114 + min(yb)]

        xt = (a - 25) * np.cos(t)  # top coordinates
        yt = (b - 20) * np.sin(t)
        # left box
        xl = [min(xb) + 25, min(xb)]
        yl = yr

        x = np.concatenate((xb, xr, xt[::-1], xl))
        y = np.concatenate((yb, yr, yt[::-1], yl))

        u_coord = zip(x, y)
        u_shape = Polygon(u_coord)

        wwu_unscaled = cascaded_union([w_shape, translate(w_shape, 139), translate(u_shape, 261, 29)])

        # create whole logo
        if self.text == 0:
            # no text
            # calculate scaling of logo and WWU for set height
            m = self.height / h  # scaling factor

            logo_complete = scale(logo_unscaled, m, m, origin=(0, 0))
        elif self.text == 1:
            # text right of logo
            # calculate scaling of logo and WWU for set height
            m = self.height / h  # scaling factor
            m2 = (y3 + h3) * m / 114

            logo_complete = cascaded_union([scale(logo_unscaled, m, m, origin=(0, 0)),
                                            translate(scale(wwu_unscaled, m2, m2, origin=(0, 0)),
                                                      (x1 + w) * m + 80 * m2)])
        else:
            # text under logo
            # calculate scaling of logo and WWU with sum equal to set height
            # WWU width = 368.93 = 369     *M2 == w
            # box width unscaled = w = 336
            m2 = w / 369.  # scale width of WWU to width of logo
            m = self.height / (h + 10 + 114)  # scaling factor height

            logo_complete_unscaled = cascaded_union(
                [scale(translate(wwu_unscaled, 67, 0), m2, m2, origin=(0, 0)), translate(logo_unscaled, 0, (114 + 10))])
            logo_complete = scale(logo_complete_unscaled, m, m, origin=(0, 0))

        return translate(logo_complete, *self.origin)


def _example():
    import gdsCAD.core
    from gdshelpers.geometry import convert_to_gdscad

    kit_logo = KITLogo([0, 0], 1)
    wwu_logo = WWULogo([0, 0], 1, 1)

    cell = gdsCAD.core.Cell('LOGOS')
    cell.add(convert_to_gdscad(kit_logo))
    cell.add(convert_to_gdscad(translate(wwu_logo.get_shapely_object(), 2.5)))
    cell.show()


if __name__ == '__main__':
    _example()



import numpy as np
import shapely.geometry
import shapely.ops
import shapely.affinity

from gdshelpers.parts import _fonts
from gdshelpers.helpers import normalize_phase
from gdshelpers.helpers.alignment import Alignment


class Text(object):
    def __init__(self, origin, height, text='', alignment='left-bottom', angle=0., font='stencil', line_spacing=1.5,
                 true_bbox_alignment=False):
        self.origin = origin
        self.height = height
        self.text = str(text)
        self._alignment = Alignment(alignment)
        self.angle = normalize_phase(angle)
        self.font = font
        self.line_spacing = height * line_spacing
        self.true_bbox_alignment = true_bbox_alignment
        self._bbox = None
        self._shapely_object = None

    def _invalidate(self):
        self._bbox = None
        self._shapely_object = None

    @property
    def origin(self):
        return self._origin

    # noinspection PyAttributeOutsideInit
    @origin.setter
    def origin(self, origin):
        self._invalidate()
        self._origin = np.array(origin)
        assert self._origin.shape == (2,), 'origin not valid'

    @property
    def height(self):
        return self._height

    # noinspection PyAttributeOutsideInit
    @height.setter
    def height(self, height):
        self._invalidate()
        assert height > 0, 'Height must be positive'
        self._height = height

    @property
    def font(self):
        return self._font

    # noinspection PyAttributeOutsideInit
    @font.setter
    def font(self, font):
        self._invalidate()
        assert font in _fonts.FONTS, 'Font is "%s" unknown, must be one of %s' % (font, _fonts.FONTS.keys())
        self._font = font

    @property
    def alignment(self):
        return self._alignment.alignment

    @alignment.setter
    def alignment(self, alignment):
        self._invalidate()
        self._alignment = alignment

    @property
    def bounding_box(self):
        if self._bbox is None:
            self.get_shapely_object()

        return self._bbox

    def get_shapely_object(self):
        if not self.text:
            self._bbox = None
            return shapely.geometry.Polygon()

        if self._shapely_object:
            return self._shapely_object

        # Let's do the actual rendering

        polygons = list()

        special_handling_chars = '\n'
        font = _fonts.FONTS[self.font]

        # Check the text
        for char in self.text:
            if char in special_handling_chars:
                continue
            assert char in font, 'Character "%s" is not supported by font "%s"' % (char, self.font)

        max_x = 0
        cursor_x, cursor_y = 0, 0
        for i, char in enumerate(self.text):
            if char == '\n':
                cursor_x, cursor_y = 0, cursor_y - self.line_spacing
                continue

            char_font = font[char]
            cursor_x += char_font['width'] / 2 * self.height

            for line in char_font['lines']:
                points = np.array(line).T * self.height + (cursor_x, cursor_y)
                polygons.append(shapely.geometry.Polygon(points))

            # Add kerning
            if i < len(self.text) - 1 and self.text[i + 1] not in special_handling_chars:
                kerning = char_font['kerning'][self.text[i + 1]]
                cursor_x += (char_font['width'] / 2 + kerning) * self.height

            max_x = max(max_x, cursor_x + char_font['width'] / 2 * self.height)

        merged_polygon = shapely.ops.cascaded_union(polygons)

        # Handle the alignment, translation and rotation
        if not self.true_bbox_alignment:
            bbox = np.array([[0, max_x],
                             [cursor_y, self.height]]).T
        else:
            bbox = np.array(merged_polygon.bounds).reshape(2, 2)

        offset = self._alignment.calculate_offset(bbox)

        if not np.isclose(normalize_phase(self.angle), 0):
            aligned_text = shapely.affinity.translate(merged_polygon, *offset)
            rotated_text = shapely.affinity.rotate(aligned_text, self.angle, origin=[0, 0], use_radians=True)
            final_text = shapely.affinity.translate(rotated_text, *self.origin)
        else:
            final_text = shapely.affinity.translate(merged_polygon, *(offset + self.origin))

        self._bbox = np.array(final_text.bounds).reshape(2, 2)
        self._shapely_object = final_text
        return final_text


def _example():
    text = Text([100, 100], 10, 'The quick brown fox jumps over the lazy dog\n123\n4567',
                alignment='left-bottom', angle=0.1)
    print(text.bounding_box)

    from gdshelpers.geometry.chip import Cell

    cell = Cell('FONTS')
    cell.add_to_layer(2, shapely.geometry.box(*text.bounding_box.reshape(4)))
    cell.add_to_layer(1, text)
    cell.show()


if __name__ == '__main__':
    _example()



from itertools import product

from gdshelpers.parts.marker import SquareMarker


def raith_marker_frame(bounds, padding=100, pitch=200, size=20, n=5):
    """
    Generates a list of markers markers in each corner around the given bounding box.
    In each corner the markers are arranged in an L shape. This allows to have more
    markers (for exposure of many steps sometimes more than four markers are necessary)
    with larger distance between the markers (minimize risk of wrong markers being
    found by the EBL) without taking up too much space.

    :param bounds: The bounds around which the markers will be arranged (the marker centers will be inside these bounds)
    :param padding: Spacing between the given bounds and the markers.
        Can also be negative to place the markers inside the bounding box.
    :param pitch: Pitch between two adjacent markers
    :param size: The marker size
    :param n: This determines the number of markers: There will be (2*n)+1 markers in each corner.
    """
    marker_offsets = [(0, 0)] + [(0, (i + 1) * pitch) for i in range(n)] + [((i + 1) * pitch, 0) for i in range(n)]
    signs = product([-1, 1], repeat=2)
    x, y = 0.5 * (bounds[0] + bounds[2]), 0.5 * (bounds[1] + bounds[3])
    half_width, half_height = 0.5 * (bounds[2] - bounds[0]) + padding, 0.5 * (bounds[3] - bounds[1]) + padding

    return [SquareMarker((x + (half_width - off[0]) * sign[0], y + (half_height - off[1]) * sign[1]), size)
            for sign, off in product(signs, marker_offsets)]



import numpy as np

from gdshelpers.parts import Port
from gdshelpers.geometry import convert_to_gdscad
from gdshelpers.parts.waveguide import Waveguide
from gdshelpers.parts.splitter import Splitter, MMI
from gdshelpers.geometry import geometric_union


class MachZehnderInterferometer(object):
    """
    A simple Mach-Zehnder interferometer based on Y-splitters.

    :param origin: Start of the interferometer.
    :type origin: tuple
    :param angle: Angle of the interferometer in rad.
    :type angle: float
    :param width: Waveguide width.
    :type width: float
    :param splitter_length: Length of the splitter
    :type splitter_length: float
    :param splitter_separation: Separation of the slitter branches.
    :type splitter_separation: float
    :param bend_radius: Bend radius.
    :type bend_radius: float
    :param upper_vertical_length: Straight length of the upper branch.
    :type upper_vertical_length: float
    :param lower_vertical_length: Straight length of the lower branch.
    :type lower_vertical_length: float
    :param horizontal_length: Straight horizontal length for both branches.
    :type horizontal_length: float
    """

    def __init__(self, origin, angle, width, splitter_length, splitter_separation, bend_radius,
                 upper_vertical_length, lower_vertical_length, horizontal_length):
        self.origin = origin
        self.angle = angle
        self.width = width
        self.splitter_length = splitter_length
        self.splitter_separation = splitter_separation
        self.bend_radius = bend_radius
        self.upper_vertical_length = upper_vertical_length
        self.lower_vertical_length = lower_vertical_length
        self.horizontal_length = horizontal_length
        self.dev_width = 2 * self.splitter_length + 4 * self.bend_radius + self.horizontal_length + 20

    @classmethod
    def make_at_port(cls, port, splitter_length, splitter_separation, bend_radius,
                     upper_vertical_length, lower_vertical_length, horizontal_length):
        return cls(port.origin, port.angle, port.width, splitter_length, splitter_separation, bend_radius,
                   upper_vertical_length, lower_vertical_length, horizontal_length)

    @property
    def port(self):
        port = Port(self.origin, self.angle, self.width)
        return port.longitudinal_offset(2 * self.splitter_length + 4 * self.bend_radius + self.horizontal_length)

    def get_shapely_object(self):
        splitter1 = Splitter(self.origin, self.angle, self.splitter_length, self.width, self.splitter_separation)

        upper_wg = Waveguide.make_at_port(splitter1.left_branch_port)
        upper_wg.add_bend(np.deg2rad(90), self.bend_radius)
        upper_wg.add_straight_segment(self.upper_vertical_length)
        upper_wg.add_bend(np.deg2rad(-90), self.bend_radius)
        upper_wg.add_straight_segment(self.horizontal_length)
        upper_wg.add_bend(np.deg2rad(-90), self.bend_radius)
        upper_wg.add_straight_segment(self.upper_vertical_length)
        upper_wg.add_bend(np.deg2rad(90), self.bend_radius)

        lower_wg = Waveguide.make_at_port(splitter1.right_branch_port)
        lower_wg.add_bend(np.deg2rad(-90), self.bend_radius)
        lower_wg.add_straight_segment(self.lower_vertical_length)
        lower_wg.add_bend(np.deg2rad(90), self.bend_radius)
        lower_wg.add_straight_segment(self.horizontal_length)
        lower_wg.add_bend(np.deg2rad(90), self.bend_radius)
        lower_wg.add_straight_segment(self.lower_vertical_length)
        lower_wg.add_bend(np.deg2rad(-90), self.bend_radius)

        splitter2 = Splitter.make_at_right_branch_port(upper_wg.current_port, self.splitter_length,
                                                       self.splitter_separation)

        return geometric_union([splitter1, splitter2, upper_wg, lower_wg])

    @property
    def device_width(self):
        return self.dev_width


class MachZehnderInterferometerMMI(object):
    """
        A simple Mach-Zehnder interferometer based on Y-splitters.

        :param origin: Start of the interferometer.
        :type origin: tuple
        :param angle: Angle of the interferometer in rad.
        :type angle: float
        :param width: Waveguide width.
        :type width: float
        :param splitter_length: Length of the splitter
        :type splitter_length: float
        :param splitter_separation: Separation of the slitter branches.
        :type splitter_separation: float
        :param bend_radius: Bend radius.
        :type bend_radius: float
        :param upper_vertical_length: Straight length of the upper branch.
        :type upper_vertical_length: float
        :param lower_vertical_length: Straight length of the lower branch.
        :type lower_vertical_length: float
        :param horizontal_length: Straight horizontal length for both branches.
        :type horizontal_length: float
    """

    def __init__(self, origin, angle, width, splitter_length, splitter_width, bend_radius,
                 upper_vertical_length, lower_vertical_length, horizontal_length):
        self.origin = origin
        self.angle = angle
        self.width = width
        self.splitter_length = splitter_length
        self.splitter_width = splitter_width
        self.bend_radius = bend_radius
        self.upper_vertical_length = upper_vertical_length
        self.lower_vertical_length = lower_vertical_length
        self.horizontal_length = horizontal_length
        self.dev_width = 2 * self.splitter_length + 4 * self.bend_radius + self.horizontal_length + 20

    @classmethod
    def make_at_port(cls, port, splitter_length, splitter_width, bend_radius,
                     upper_vertical_length, lower_vertical_length, horizontal_length):
        return cls(port.origin, port.angle, port.width, splitter_length, splitter_width, bend_radius,
                   upper_vertical_length, lower_vertical_length, horizontal_length)

    @property
    def port(self):
        port = Port(self.origin, self.angle, self.width)
        return port.longitudinal_offset(2 * self.splitter_length + 4 * self.bend_radius + self.horizontal_length + 20)

    @property
    def device_width(self):
        return self.dev_width

    def get_shapely_object(self):
        splitter1 = MMI(origin=self.origin, angle=self.angle, wg_width=self.width, length=self.splitter_length,
                        width=self.splitter_width, num_inputs=1, num_outputs=2)
        print(splitter1.output_ports)
        upper_wg = Waveguide.make_at_port(splitter1.left_branch_port)
        upper_wg.add_bend(np.deg2rad(90), self.bend_radius)
        upper_wg.add_straight_segment(self.upper_vertical_length)
        upper_wg.add_bend(np.deg2rad(-90), self.bend_radius)
        upper_wg.add_straight_segment(self.horizontal_length)
        upper_wg.add_bend(np.deg2rad(-90), self.bend_radius)
        upper_wg.add_straight_segment(self.upper_vertical_length)
        upper_wg.add_bend(np.deg2rad(90), self.bend_radius)

        lower_wg = Waveguide.make_at_port(splitter1.right_branch_port)
        lower_wg.add_bend(np.deg2rad(-90), self.bend_radius)
        lower_wg.add_straight_segment(self.lower_vertical_length)
        lower_wg.add_bend(np.deg2rad(90), self.bend_radius)
        lower_wg.add_straight_segment(self.horizontal_length)
        lower_wg.add_bend(np.deg2rad(90), self.bend_radius)
        lower_wg.add_straight_segment(self.lower_vertical_length)
        lower_wg.add_bend(np.deg2rad(-90), self.bend_radius)

        splitter2 = MMI(origin=[self.origin[0] + self.dev_width, self.origin[1]], angle=self.angle + np.pi,
                        wg_width=self.width, length=self.splitter_length,
                        width=self.splitter_width, num_inputs=1, num_outputs=2)
        print(splitter2.input_ports)

        return geometric_union([splitter1, splitter2, upper_wg, lower_wg])


def _example():
    import gdsCAD

    devicename = 'MZI'

    mzi = MachZehnderInterferometer(origin=(0, 0), angle=0, width=1.2, splitter_length=10, splitter_separation=5,
                                    bend_radius=50, upper_vertical_length=50, lower_vertical_length=0,
                                    horizontal_length=0)
    mzi_mmi = MachZehnderInterferometerMMI(origin=(300, 0), angle=0, width=1.2, splitter_length=33, splitter_width=7.7,
                                           bend_radius=50, upper_vertical_length=50, lower_vertical_length=0,
                                           horizontal_length=0)
    print(mzi_mmi.device_width)

    wg = Waveguide.make_at_port(mzi_mmi.port)
    wg.add_straight_segment(length=50)

    cell = gdsCAD.core.Cell(devicename)
    cell.add(convert_to_gdscad(mzi, layer=1))
    cell.add(convert_to_gdscad(mzi_mmi, layer=1))
    cell.add(convert_to_gdscad(wg, layer=1))

    layout = gdsCAD.core.Layout()
    layout.add(cell=cell)
    layout.save('%s.gds' % devicename)
    layout.show()


if __name__ == '__main__':
    _example()



from shapely.geometry.polygon import Polygon
from shapely.geometry.multipolygon import MultiPolygon
from shapely.affinity import translate, rotate, scale
import gdspy


class GDSIIImport:
    def __init__(self, filename, cell_name, layer=None, datatype=None):
        self.gdslib = gdspy.GdsLibrary(infile=filename)
        self.cell_name = cell_name
        self.layer = layer
        self.datatype = datatype

    def get_as_shapely(self, cell, layer=None, datatype=None):
        geometry = []

        gdspy_cell = self.gdslib.cell_dict[cell] if isinstance(cell, str) else cell
        for polygon in gdspy_cell.polygons:
            if self.layer is not None and layer != polygon.layers[0]:
                continue
            if self.datatype is not None and datatype != polygon.datatypes[0]:
                continue
            geometry.append(Polygon(polygon.polygons[0]).buffer(0))  # .buffer(0) for healing geometries

        for reference in gdspy_cell.references:
            sub_geometry = self.get_as_shapely(reference.ref_cell, layer, datatype)
            if sub_geometry.is_empty:
                continue
            sub_geometry = scale(sub_geometry,
                                 *[reference.magnification] * 2) if reference.magnification else sub_geometry
            sub_geometry = scale(sub_geometry, -1) if reference.x_reflection else sub_geometry
            sub_geometry = rotate(sub_geometry, reference.rotation,
                                  origin=(0, 0)) if reference.rotation else sub_geometry
            sub_geometry = translate(sub_geometry, *reference.origin)
            geometry.extend(sub_geometry)

        return MultiPolygon(geometry)

    def get_shapely_object(self):
        return self.get_as_shapely(self.cell_name, self.layer, self.datatype)


if __name__ == '__main__':
    from gdshelpers.parts.waveguide import _example
    from gdshelpers.geometry.chip import Cell

    _example()  # make sure the "output.gds" exists

    test_cell = Cell('test')
    test_cell.add_to_layer(1, GDSIIImport('output.gds', cell_name='TOP', layer=1))
    test_cell.show()



import unittest

from gdshelpers.parts.splitter import Splitter


class SplitterTestCase(unittest.TestCase):
    def test_splitter_same_wg_width(self):
        total_length = 10
        wg_width = 1
        sep = 5

        splitter = Splitter([0, 0], 0, total_length, wg_width, sep)
        self.assertAlmostEqual(splitter.root_port.width, wg_width)
        self.assertAlmostEqual(splitter.left_branch_port.width, wg_width)
        self.assertAlmostEqual(splitter.left_branch_port.width, wg_width)

        self.assertAlmostEqual(splitter.root_port.origin[0], 0)
        self.assertAlmostEqual(splitter.root_port.origin[1], 0)



import unittest
import numpy as np
import numpy.testing as npt

from gdshelpers.helpers import int_to_alphabet, id_to_alphanumeric, find_line_intersection


class HelpersTestCase(unittest.TestCase):
    def test_int_to_alphabet(self):
        self.assertEqual(int_to_alphabet(0), 'A')
        self.assertEqual(int_to_alphabet(25), 'Z')
        self.assertEqual(int_to_alphabet(26), 'AA')
        self.assertEqual(int_to_alphabet(26 + 25), 'AZ')

    def test_id_to_alphanumeric(self):
        self.assertEqual(id_to_alphanumeric(4, 26 + 25), 'AZ4')

    def test_find_line_intersection(self):
        test_intersection = find_line_intersection(np.array((2, 0)), np.pi / 2, np.array((0, 1)), 0)
        npt.assert_almost_equal(test_intersection[0], (2, 1))
        npt.assert_almost_equal(test_intersection[1], (1, 2))



import numpy as np

from gdshelpers.geometry import geometric_union
from gdshelpers.parts import Port
from gdshelpers.parts.waveguide import Waveguide


class Spiral:
    def __init__(self, origin, angle, width, num, gap, inner_gap):
        """
        Creates a Spiral around the given origin

        :param origin: position of the center of the spiral
        :param angle: angle of the outer two waveguides
        :param width: width of the waveguide
        :param num: number of turns
        :param gap: gap between two waveguides
        :param inner_gap: inner radius of the spiral
        """
        self._origin_port = Port(origin, angle, width)
        self.gap = gap
        self.inner_gap = inner_gap
        self.num = num
        self.wg_in = None
        self.wg_out = None

    @classmethod
    def make_at_port(cls, port, num, gap, inner_gap):
        """
        Creates a Spiral around the given port

        :param port: port at which the spiral starts
        :param num: number of turns
        :param gap: gap between two waveguides
        :param inner_gap: inner radius of the spiral
        """
        return cls(port.parallel_offset(-num * (port.total_width + gap) - inner_gap).origin,
                   port.angle, port.width, num, gap, inner_gap)

    ###
    # Let's allow the user to change the values
    # hidden in _origin_port. Hence the internal use
    # of a Port is transparent.
    @property
    def origin(self):
        return self._origin_port.origin

    @origin.setter
    def origin(self, origin):
        self._origin_port.origin = origin

    @property
    def angle(self):
        return self._origin_port.angle

    @angle.setter
    def angle(self, angle):
        self._origin_port.angle = angle

    @property
    def width(self):
        return self._origin_port.width

    @width.setter
    def width(self, width):
        self._origin_port.width = width

    @property
    def in_port(self):
        return self._origin_port.inverted_direction.parallel_offset(
            -self.num * (self._origin_port.total_width + self.gap) - self.inner_gap)

    @property
    def out_port(self):
        return self._origin_port.parallel_offset(
            -self.num * (self._origin_port.total_width + self.gap) - self.inner_gap)

    @property
    def length(self):
        if not self.wg_in or not self.wg_out:
            self._generate()
        return self.wg_in.length + self.wg_out.length

    def _generate(self):
        def path(a):
            return (self.num * (self._origin_port.total_width + self.gap) * np.abs(1 - a) + self.inner_gap) * np.array(
                (np.sin(np.pi * a * self.num), np.cos(np.pi * a * self.num)))

        self.wg_in = Waveguide.make_at_port(self._origin_port)
        self.wg_in.add_parameterized_path(path)

        self.wg_out = Waveguide.make_at_port(self._origin_port.inverted_direction)
        self.wg_out.add_parameterized_path(path)

        self.wg_in.add_route_single_circle_to_port(self._origin_port.rotated(-np.pi * (self.num % 2)))
        self.wg_in.add_route_single_circle_to_port(self.wg_out.port)

    def get_shapely_object(self):
        if not self.wg_in or not self.wg_out:
            self._generate()
        return geometric_union([self.wg_in, self.wg_out])


def _example():
    from gdshelpers.geometry.chip import Cell

    wg = Waveguide((0, 0), 1, 1)
    wg.add_straight_segment(30)
    spiral = Spiral.make_at_port(wg.current_port, 2, 5, 50)
    print(spiral.out_port.origin)
    wg2 = Waveguide.make_at_port(spiral.out_port)
    wg2.add_straight_segment(100)

    print(spiral.length)

    cell = Cell('Spiral')
    cell.add_to_layer(1, wg, spiral, wg2)
    cell.show()


if __name__ == '__main__':
    _example()



import numpy as np

from gdshelpers.geometry import geometric_union
from gdshelpers.parts import Port
from gdshelpers.parts.waveguide import Waveguide


class SNSPD:
    def __init__(self, origin, angle, width, nano_wire_width, nano_wire_gap, nano_wire_length, waveguide_tapering,
                 passivation_buffer):
        if nano_wire_width <= 0:
            raise ValueError("The nano-wire width must be greater than 0")
        if nano_wire_length < 0:
            raise ValueError("The nano-wire length must be positive")
        if nano_wire_gap <= 0:
            raise ValueError("The nano-wire gap must be greater than 0")
        if nano_wire_length < 0:
            raise ValueError("The distance for the passivation layer must be positive")

        self._origin_port = Port(origin, angle, width)
        self.nano_wire_gap = nano_wire_gap
        self.nano_wire_width = nano_wire_width
        self.nano_wire_length = nano_wire_length
        self.waveguide_tapering = waveguide_tapering
        self.passivation_buffer = passivation_buffer

        self._waveguide = None
        self._nano_wire = None
        self._wings = None

    @classmethod
    def make_at_port(cls, port, nw_width, nw_gap, nw_length, waveguide_tapering, passivation_buffer):
        return cls(port.origin, port.angle, port.width, nw_width, nw_gap, nw_length, waveguide_tapering,
                   passivation_buffer)

    def _generate(self):
        tip_radius = 0.5 * (self.nano_wire_gap + self.nano_wire_width)
        nano_wire_port = self._origin_port \
            .longitudinal_offset(self.nano_wire_length + tip_radius + self.passivation_buffer) \
            .parallel_offset(tip_radius)

        self._nano_wire = Waveguide.make_at_port(nano_wire_port.inverted_direction, width=self.nano_wire_width)
        self._nano_wire.add_straight_segment(self.nano_wire_length - tip_radius)
        self._nano_wire.add_bend(np.pi, tip_radius)
        self._nano_wire.add_straight_segment(self.nano_wire_length - tip_radius)

        self._wings = []
        for sign, port in [(1, self._nano_wire.current_port), (-1, self._nano_wire.in_port)]:
            wing_outer_port = port.longitudinal_offset(14).parallel_offset(-sign * 10).rotated(sign * 0.75 * np.pi)
            wing_outer_port.width = 5
            wing_inner_port = wing_outer_port.longitudinal_offset(10)
            wing_inner_port.width = 0.5

            self._wings.append(Waveguide.make_at_port(port)
                               .add_bezier_to_port(port=wing_inner_port, bend_strength=2.5)
                               .add_bezier_to_port(port=wing_outer_port, bend_strength=6)
                               .add_straight_segment(1))

        self._waveguide = Waveguide.make_at_port(self._origin_port)
        self._waveguide.add_straight_segment_until_level_of_port(self._wings[0].current_port.rotated(-0.75 * np.pi))
        if self.waveguide_tapering:
            self._waveguide.add_straight_segment(5. * self._origin_port.width, final_width=0.01)

    def get_shapely_object(self):
        if not self._nano_wire or not self._wings:
            self._generate()
        return geometric_union([self._nano_wire] + self._wings)

    def get_waveguide(self):
        if not self._waveguide or not self._wings:
            self._generate()
        return geometric_union([self._waveguide] + [wing.get_shapely_object().buffer(.2) for wing in self._wings])

    def get_passivation_layer(self):
        if not self._nano_wire or not self._wings:
            self._generate()
        return geometric_union([self._nano_wire] + self._wings).buffer(self.passivation_buffer)

    @property
    def right_electrode_port(self):
        if not self._wings:
            self._generate()
        return self._wings[0].current_port.longitudinal_offset(-1)

    @property
    def left_electrode_port(self):
        if not self._wings:
            self._generate()
        return self._wings[1].current_port.longitudinal_offset(-1)

    @property
    def current_port(self):
        if not self._waveguide:
            self._generate()
        return self._waveguide.current_port


def _example():
    from gdshelpers.geometry.chip import Cell

    cell = Cell('test')

    wg1 = Waveguide((0, 0), 0.5 * np.pi, 1.)
    wg1.add_straight_segment(10.)
    cell.add_to_layer(3, wg1)

    detector = SNSPD.make_at_port(wg1.current_port, nw_width=0.1, nw_gap=0.1, nw_length=70, passivation_buffer=0.2,
                                  waveguide_tapering=True)
    cell.add_to_layer(1, detector)
    cell.add_to_layer(2, detector.get_waveguide())
    cell.add_to_layer(5, detector.get_passivation_layer())

    # cell.add_to_layer(6, detector.right_electrode_port.debug_shape)
    # cell.add_to_layer(6, detector.left_electrode_port.debug_shape)
    wg2 = Waveguide.make_at_port(detector.current_port)
    wg2.add_straight_segment(20.)
    cell.add_to_layer(3, wg2)
    cell.save('SNSPD_test.gds')
    cell.show()


if __name__ == '__main__':
    _example()



# noinspection PyUnresolvedReferences
import bpy
# noinspection PyUnresolvedReferences
import bmesh
import sys
import os
import numpy as np

# Remove the standard cube and the lamp of blender
blender_objects = bpy.data.objects
blender_objects.remove(blender_objects["Cube"], True)
blender_objects.remove(blender_objects["Lamp"], True)

scene = bpy.context.scene

# Load the 3d objects
argv = sys.argv
argv = argv[argv.index("--") + 1:]
raw_input_from_file = open(argv[0] + ".tmp").read()

for end_string in [".tmp", ".png", ".blend"]:
    if os.path.exists(argv[0] + end_string):
        os.remove(argv[0] + end_string)

blocks = raw_input_from_file.split('\n')
config = blocks[0].split(';')

index = 1
material_dict = {}

# Saves the meshes and the merge index
meshes = []
meshes_index = []

merge_index = -1

# Buffer variables needed for the calculation of the camera position
x_min = x_max = y_min = y_max = z_min = z_max = 0

# Create the meshes and link them to the scene
while index < len(blocks) - 1:
    # Read information about the mesh object
    info = blocks[index].split(';')
    merge_index = int(info[0])
    block_len = int(info[1])
    extrude_val_min = float(info[2])
    extrude_val_max = float(info[3])
    scale = float(info[4])
    rgb = [int(info[5]), int(info[6]), int(info[7])]
    index += 1

    # Read the vertices
    vertices = [(float(blocks[i].split()[0]) * scale, float(blocks[i].split()[1]) * scale, extrude_val_min * scale)
                for i in range(index, index + block_len)]

    # Have a look at the dimensions of the objects -> to set the camera position
    for vertex in vertices:
        if vertex[0] > x_max:
            x_max = vertex[0]
        if vertex[0] < x_min:
            x_min = vertex[0]
        if vertex[1] > y_max:
            y_max = vertex[1]
        if vertex[1] < y_min:
            y_min = vertex[1]

    if extrude_val_min < z_min:
        z_min = extrude_val_min
    if extrude_val_max > z_max:
        z_max = extrude_val_max

    index += block_len

    # Create the mesh, color and extrude it
    mesh = bpy.data.meshes.new("mesh_" + str(index))
    b_mesh = bmesh.new()
    b_vertices = [b_mesh.verts.new(vertex) for vertex in vertices]
    b_face = b_mesh.faces.new(b_vertices)
    b_face.normal = (0, 0, -1)
    bmesh.ops.solidify(b_mesh, geom=[b_face], thickness=extrude_val_max * scale)
    b_mesh.to_mesh(mesh)
    mesh_object = bpy.data.objects.new("obj_" + str(index), mesh)
    material_name = str(rgb[0]).ljust(3) + str(rgb[1]).ljust(3) + str(rgb[2]).ljust(3)
    structure_material = bpy.data.materials.new(material_name)
    structure_material.diffuse_color = (rgb[0] / 255, rgb[1] / 255, rgb[2] / 255)
    if material_name not in material_dict.keys():
        material_dict.update({material_name: structure_material})
    mesh_object.data.materials.append(material_dict[material_name])
    # Add to the mesh lists
    meshes_index.append(merge_index)
    meshes.append(mesh_object)
    scene.objects.link(mesh_object)


#
# Join the meshes corresponding to a specific Shapely3D object
#

def get_first_of_equal_index(index_list):
    buffer_list = []

    for test_index in index_list:
        if test_index in buffer_list:
            for j, ind in enumerate(buffer_list):
                if ind == test_index and ind != -1:
                    return j
        buffer_list.append(test_index)
    return -1


def join_on_scene(mesh_list, index_list):
    fuse = get_first_of_equal_index(index_list)
    new_list = list(index_list)

    if fuse != -1:
        bpy.ops.object.select_all(action='DESELECT')
        mesh_list[fuse].select = True
        bpy.context.scene.objects.active = mesh_list[fuse]
        for k in range(fuse + 1, len(new_list)):
            if new_list[fuse] == new_list[k]:
                mesh_list[k].select = True
                bpy.context.scene.objects.active = mesh_list[k]
                new_list[k] = -1
        new_list[fuse] = -1
        bpy.ops.object.join()
        join_on_scene(mesh_list, new_list)


join_on_scene(meshes, meshes_index)

#
# Do the calculation for the camera position
#

# The diagonal line of the floor follows the equation y(x) = m * x + b
if config[7] == 'right':
    m = (y_max - y_min) / (x_max - x_min) if config[6] == 'above' else - (y_max - y_min) / (x_max - x_min)
    b = y_min - m * x_min if config[6] == 'above' else y_min - m * x_max
else:
    m = (y_max - y_min) / (x_max - x_min) if config[6] == 'under' else - (y_max - y_min) / (x_max - x_min)
    b = y_min - m * x_min if config[6] == 'under' else y_min - m * x_max

# Length of the diagonal line
m_width = np.sqrt((x_max - x_min) ** 2 + (y_max - y_min) ** 2)
ratio = 2.5  # Define an arbitrary value which sets the distance between the camera and the objects

alpha = np.arctan(m)
m_width_cam = m_width * ratio

cam_pos_x = np.cos(alpha) * m_width_cam if config[7] == 'right' else - np.cos(alpha) * m_width_cam
cam_pos_y = m * cam_pos_x + b
cam_pos_z = m_width_cam / 2

# Set the point on the floor at which the camera will look
mid_x = (x_max + x_min) / 2
mid_y = (y_max + y_min) / 2

rot_y = 0
rot_x = 90 - 180 / np.pi * np.arctan(cam_pos_z / np.sqrt((cam_pos_x - mid_x) ** 2 + (cam_pos_y - mid_y) ** 2))

if config[6] == 'above':
    rot_z = np.abs(180 / np.pi * np.arctan(m)) + 90 if config[7] == 'right' else np.abs(
        180 / np.pi * np.arctan(m)) + 225
else:
    rot_z = np.abs(180 / np.pi * np.arctan(m)) + 45 if config[7] == 'right' else np.abs(
        180 / np.pi * np.arctan(m)) + 270

#
# Add the new lamp to the scene and set the calculated camera parameters
#

lamp_data = bpy.data.lamps.new(name="New Lamp", type='SUN')
lamp_object = bpy.data.objects.new(name="New Lamp", object_data=lamp_data)
scene.objects.link(lamp_object)
lamp_object.location = (cam_pos_x, cam_pos_y, cam_pos_z)  # Place the lamp exactly on the camera

lamp_object.data.use_nodes = True
lamp_object.data.node_tree.nodes['Emission'].inputs['Strength'].default_value = 1

scene.camera.location.x = cam_pos_x
scene.camera.location.y = cam_pos_y
scene.camera.location.z = cam_pos_z

scene.camera.rotation_mode = 'XYZ'
scene.camera.rotation_euler[0] = rot_x * (np.pi / 180.0)
scene.camera.rotation_euler[1] = rot_y * (np.pi / 180.0)
scene.camera.rotation_euler[2] = rot_z * (np.pi / 180.0)

scene.camera.data.clip_end = 500

bpy.data.scenes["Scene"].render.resolution_x = int(config[2])
bpy.data.scenes["Scene"].render.resolution_y = int(config[3])
bpy.data.scenes["Scene"].render.resolution_percentage = int(config[4])
bpy.context.scene.render.engine = config[5]
bpy.context.scene.cycles.samples = 128
if config[0] == '1':
    bpy.data.scenes["Scene"].render.filepath = sys.argv[3].rsplit('\\', 1)[0] + '\\' + argv[0] + '.png'
    bpy.ops.render.render(write_still=True)
if config[1] == '1':
    bpy.ops.wm.save_mainfile(filepath=argv[0] + ".blend")



import math
import numpy as np

from gdshelpers.helpers import normalize_phase


class Port(object):
    """
    Abstraction of a waveguide port.

    Other objects might dock to a port. It is simply a helper object
    to allow easy chaining of parts.

    :param origin: Origin of the port.
    :param angle: Angle of the port.
    :param width: Width of the port.
    :type width: float
    """

    def __init__(self, origin, angle, width):
        self.origin = origin
        self.angle = angle
        self.width = width

    def copy(self):
        """
        Create a copy if the port.

        :return: A copy of the port.
        :rtype: Port
        """
        return Port(self.origin, self.angle, self.width)

    def __getitem__(self, item):
        if isinstance(item, slice):
            start = item.start if item.start >= 0 else len(self.width[::2]) + item.start + 1
            stop = item.stop if item.stop >= 0 else len(self.width[::2]) + item.stop + 1
            start, stop = 2 * start, 2 * stop
        else:
            item = item if item >= 0 else len(self.width[::2]) + item
            start, stop = 2 * item, 2 * item + 1

        port = self.parallel_offset((np.sum(self.width[:stop]) + np.sum(self.width[:start])) / 2 - self.total_width / 2)
        port.width = self.width[start:stop]
        return port

    def get_parameters(self):
        """
        Get a dictionary representation of the port properties.

        :return: A dictionary containing the ``origin``, ``angle`` and ``width`` of the port.
        :rtype: dict
        """
        return {key: getattr(self, key) for key in ('origin', 'angle', 'width')}

    def set_port_properties(self, **kwargs):
        """
        Set port parameters via named keyword arguments.

        :param kwargs: The keywords to set.

        :return: The modified port
        :rtype: Port
        """
        for key, value in kwargs.items():
            assert key in ('origin', 'angle', 'width'), '"%s" is not a valid property' % key
            setattr(self, key, value)

        return self

    @property
    def inverted_direction(self):
        """
        Get a port which points in the opposite direction.

        :return: A copy of this port, pointing in the opposite direction.
        :rtype: Port
        """
        inverted_port = self.copy()
        inverted_port.angle = inverted_port.angle + math.pi
        return inverted_port

    @property
    def origin(self):
        """
        The origin coordinates of this port.

        When reading it is guarantied to be a 2-dim numpy array.
        """
        return self._origin

    # noinspection PyAttributeOutsideInit
    @origin.setter
    def origin(self, origin):
        assert len(origin) == 2, 'origin must be a 2D coordinate'
        self._origin = np.array(origin, dtype=float)

    @property
    def x(self):
        return self._origin[0]

    @x.setter
    def x(self, x):
        self._origin[0] = x

    @property
    def y(self):
        return self._origin[1]

    @y.setter
    def y(self, y):
        self._origin[1] = y

    @property
    def angle(self):
        """
        The angle of the port.
        """
        return normalize_phase(self._angle)

    # noinspection PyAttributeOutsideInit
    @angle.setter
    def angle(self, angle):
        self._angle = angle % (2 * math.pi)

    @property
    def width(self):
        """
        The width of the port. E.g. for slot waveguides it can also be an array in the format [width, gap, width, ...],
        where each width describe the width of each rail and the gap defines the gap in between.
        This array can also end with a gap, which facilitates e.g. the design of adiabatic mode converters.
        """
        return self._width

    @property
    def total_width(self):
        """
        The total width of the port.

        Guarantied to be a positive float.
        """
        return np.sum(self._width)

    # noinspection PyAttributeOutsideInit
    @width.setter
    def width(self, width):
        assert np.sum(width) > 0, 'Port width must be larger than zero'
        self._width = np.array(width) if np.array(width).size > 1 else float(width)

    def parallel_offset(self, offset):
        """
        Returns a new port, which offset in parallel from this port.

        :param offset: Offset from the center of the port. Positive is left of the port.
        :type offset: float
        :return: The new offset port
        :rtype: Port
        """
        port = self.copy()
        offset = [offset * np.cos(self.angle + np.pi / 2), offset * np.sin(self.angle + np.pi / 2)]
        port.origin = port.origin + offset
        return port

    def longitudinal_offset(self, offset):
        """
        Returns a new port, which offset in in direction of this port.

        :param offset: Offset from the end of the port. Positive is the direction, the port is pointing.
        :type offset: float
        :return: The new offset port
        :rtype: Port
        """

        port = self.copy()
        offset = [offset * np.cos(self.angle), offset * np.sin(self.angle)]
        port.origin = port.origin + offset
        return port

    def rotated(self, angle):
        """
        Returns a new port, which is rotated by the given angle.

        :param angle: Angle to rotate.
        :type angle: float
        :return: The new rotated port
        :rtype: Port
        """

        port = self.copy()
        port.angle += angle
        return port

    @property
    def debug_shape(self):
        from gdshelpers.parts.waveguide import Waveguide
        d = self.total_width / 5
        wg = Waveguide.make_at_port(self.longitudinal_offset(-d), width=self.total_width * 10)
        wg.add_straight_segment(d)
        wg.width = self.width
        wg.add_straight_segment(4 * self.total_width)
        wg.width = self.width * 5
        wg.add_straight_segment(self.total_width * 5, final_width=self.total_width * 0.1)
        return wg.get_shapely_object()



"""
Ring and race track resonators
"""

from __future__ import print_function, division

import math

import numpy as np

from gdshelpers.geometry import geometric_union
from gdshelpers.parts import Port
from gdshelpers.parts.waveguide import Waveguide


class RingResonator(object):
    """
    A simple Ring / Race track resonator.

    This part implements a super simple ring resonator with optional race tracks. Several
    helper functions are available to calculate points and ports of interest. The width of the
    feeding waveguide and the ring waveguide may differ.


    :param origin: Origin of the resonator, which is the start of the input waveguide.
    :param angle: Angle of the input waveguide.
    :param width: Width of the angle waveguide.
    :param gap: Gap between ring and waveguide. If positive, the ring will be on the left, and on the right side for
                negative gap values. Can also be a 2-tuple, if input and output gap should be different.
    :param radius: Radius of the bends.
    :param race_length: Length of the race track. Defaults to zero.
    :param draw_opposite_side_wg: If True, draw the opposing waveguides, (a.k.a. drop ports.)
    :param res_wg_width: Width of the resonator waveguide. If None, the width of the input waveguide is assumend.
    :param n_points: Number of points used per quarter circle. If None, it uses the bend default.
    :param straight_feeding: Add straight connections on both sides of the resonator.
    :param vertical_race_length: Length of a vertical race track section. Defaults to zero.
    """

    def __init__(self, origin, angle, width, gap, radius, race_length=0, draw_opposite_side_wg=False,
                 res_wg_width=None, n_points=None, straight_feeding=False, vertical_race_length=0):
        assert race_length >= 0, 'The race track length must not be negative'
        assert vertical_race_length >= 0, 'The vertical race track length must not be negative'
        assert radius > 0, 'The bend radius must not be negative'

        # Let's directly create a port object. This simplifies later creation of the geometry
        # as well as checking the user parameters while creating the port.
        self._origin_port = Port(origin, angle, width)

        # Ring gap
        try:
            self.gap, self.opposite_gap = gap
            if np.sign(self.gap) != np.sign(self.opposite_gap):
                raise ValueError
        except TypeError:
            self.gap = gap
            self.opposite_gap = gap

        # Remember all the other stuff we got
        self.radius = radius
        self.race_length = race_length
        self.res_wg_width = res_wg_width if res_wg_width else width
        self.points = n_points
        self.draw_opposite_side_wg = draw_opposite_side_wg
        self.straight_feeding = straight_feeding
        self.vertical_race_length = vertical_race_length

    @classmethod
    def make_at_port(cls, port, gap, radius, **kwargs):
        default_port_param = dict(port.get_parameters())
        default_port_param.update(kwargs)
        del default_port_param['origin']
        del default_port_param['angle']
        del default_port_param['width']

        return cls(port.origin, port.angle, port.width, gap, radius, **default_port_param)

    ###
    # Let's allow the user to change the values
    # hidden in _origin_port. Hence the internal use
    # of a Port is transparent.
    @property
    def origin(self):
        return self._origin_port.origin

    @origin.setter
    def origin(self, origin):
        self._origin_port.origin = origin

    @property
    def angle(self):
        return self._origin_port.angle

    @angle.setter
    def angle(self, angle):
        self._origin_port.angle = angle

    @property
    def width(self):
        return self._origin_port.width

    @width.setter
    def width(self, width):
        self._origin_port.width = width

    ####
    # Now, lets get to the functions the user is actually
    # interested in.
    @property
    def port(self):
        offset = self.race_length + 2 * self.radius if self.straight_feeding else 0
        return self._origin_port.longitudinal_offset(offset)

    out_port = port

    @property
    def opposite_side_port_out(self):
        return self.port.parallel_offset(np.copysign(2 * self.radius + self.vertical_race_length, self.opposite_gap)
                                         + self._offset + self._offset_opposite)

    @property
    def opposite_side_port_in(self):
        return self._origin_port.parallel_offset(
            np.copysign(2 * self.radius + self.vertical_race_length, self.opposite_gap)
            + self._offset + self._offset_opposite).inverted_direction

    # Conventional naming of ports
    @property
    def in_port(self):
        return self._origin_port.inverted_direction

    @property
    def add_port(self):
        return self.opposite_side_port_out

    @property
    def drop_port(self):
        return self.opposite_side_port_in

    @property
    def through_port(self):
        return self.port

    @property
    def center_coordinates(self):
        return self._origin_port.longitudinal_offset(self.race_length / 2.).parallel_offset(
            self._offset + np.copysign(self.radius + 0.5 * self.vertical_race_length, self.gap)).origin

    @property
    def _offset(self):
        return math.copysign(abs(self.gap) + (self.width + self.res_wg_width) / 2., self.gap)

    @property
    def _offset_opposite(self):
        return math.copysign(abs(self.opposite_gap) + (self.width + self.res_wg_width) / 2., self.opposite_gap)

    @property
    def circumference(self):
        return 2 * np.pi * self.radius + 2 * self.race_length

    def get_shapely_object(self):
        wg = Waveguide.make_at_port(self._origin_port)
        opposite_wg = Waveguide.make_at_port(self.opposite_side_port_in.inverted_direction)

        if self.straight_feeding:
            wg.add_straight_segment(self.radius)
            if self.draw_opposite_side_wg:
                opposite_wg.add_straight_segment(self.radius)

        ring_port = wg.current_port.parallel_offset(self._offset)
        ring_port.width = self.res_wg_width
        ring = Waveguide.make_at_port(ring_port)

        if self.race_length:
            wg.add_straight_segment(self.race_length)

            if self.draw_opposite_side_wg:
                opposite_wg.add_straight_segment(self.race_length)

        if self.straight_feeding:
            wg.add_straight_segment(self.radius)
            if self.draw_opposite_side_wg:
                opposite_wg.add_straight_segment(self.radius)

        # Build the ring
        bend_angle = math.copysign(0.5 * np.pi, self.gap)
        if self.race_length:
            ring.add_straight_segment(self.race_length)

        ring.add_bend(bend_angle, self.radius, n_points=self.points)
        if self.vertical_race_length:
            ring.add_straight_segment(self.vertical_race_length)

        ring.add_bend(bend_angle, self.radius, n_points=self.points)
        if self.race_length:
            ring.add_straight_segment(self.race_length)

        ring.add_bend(bend_angle, self.radius, n_points=self.points)
        if self.vertical_race_length:
            ring.add_straight_segment(self.vertical_race_length)

        ring.add_bend(bend_angle, self.radius, n_points=self.points)

        return geometric_union([ring, wg, opposite_wg])


def _example():
    from gdshelpers.geometry import convert_to_gdscad
    import gdsCAD

    cell = gdsCAD.core.Cell('test')

    wg1 = Waveguide.make_at_port(Port((0, 0), 0, 1.))
    wg1.add_straight_segment(100)
    ring1 = RingResonator.make_at_port(wg1.current_port, 1., 50., race_length=30, straight_feeding=True,
                                       draw_opposite_side_wg=True)

    wg2 = Waveguide.make_at_port(ring1.port)
    wg2.add_straight_segment(100)
    ring2 = RingResonator.make_at_port(wg2.current_port, 1., 50., vertical_race_length=30, straight_feeding=True,
                                       draw_opposite_side_wg=True)

    wg3 = Waveguide.make_at_port(ring2.port)
    wg3.add_straight_segment(100)
    ring3 = RingResonator.make_at_port(wg3.current_port, -1., 50., vertical_race_length=30, straight_feeding=True,
                                       draw_opposite_side_wg=True)

    cell.add(convert_to_gdscad([wg1, ring1, wg2, ring2, wg3, ring3], layer=1))
    cell.show()


if __name__ == '__main__':
    _example()
